/*
 * Copyright 2001 Jesus Salvo
 */
/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * BFMultiStartDirectedGraphTraversal.java
 * 
 * Time-stamp: <02/05/01 20:43:31 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: BFMultiStartDirectedGraphTraversal.java,v 1.3 2004/10/27 17:16:52 ftaiani Exp $
 * 
 */
package laas.friends.cosmopen.graphAnalysis  ;

import salvo.jesus.graph.algorithm.*;
import salvo.jesus.graph.*;
import java.util.*;
// import salvo.jesus.util.*;

/**
 * This class is directly adapted from the code of 
 * {@link salvo.jesus.graph.algorithm.BreadthFirstTraversal }
 * to allow the traversal of a directed graph while starting
 * from a set of point instead of from only one.
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.3 $
 * @since Version 0.0.3
 */
public class BFMultiStartDirectedGraphTraversal {
    List                  list     ;
    DirectedGraphWithDict dgraph    ;
    private boolean       _isDirect ; // Traversal direction (direct / reverse)

    static final public int   TERMINATEDBYVISITOR = -1;
    static final public int   OK = 1;

    /**
     * Creates a DepthFirstDirectedGraphTraversal object that will perform
     * a breath first traversal on the specified DirectedGraphWithDict. This
     * traversal can be either 'direct' or 'reverse' w.r.t. to the
     * direction of the edges. (I.e. with reverse, edges are gone along
     * from the sink to the source.)<p>
     * This traversal object is particular because it allows a traversal
     * to start from multiple point, thus allowing the computation
     * of forward / backward closures for clouds of nodes.<p>
     *
     * @param   graph    DirectedGraphWithDict on which the 
     *                   traversal will be performed.
     * @param   isDirect use <code>true</code> for a direct traversal and
     *                   <code>false</code> a reverse one.
     */
    public BFMultiStartDirectedGraphTraversal
	(  DirectedGraphWithDict graph, 
	   boolean               isDirect ) 
    {
	this.dgraph    = (DirectedGraphWithDict) graph;
	this.list      = new LinkedList() ;
	this._isDirect = isDirect ;
    }

    /**
     * Same as 
     * {@link #BFMultiStartDirectedGraphTraversal(DirectedGraphWithDict,boolean)}
     * but with a limit on the maximum recursion depth during the
     * breath first traversal. (<code>aDepthLimit</code>)
     */
    public BFMultiStartDirectedGraphTraversal
	(  DirectedGraphWithDict graph, 
	   boolean               isDirect,
	   int                   aDepthLimit ) 
    {
	this.dgraph    = (DirectedGraphWithDict) graph;
	this.list      = new LinkedList();
	this._isDirect = isDirect ;
	this._depthLimit = aDepthLimit ;
    }

    /**
     * This inner class is used in {@link #traverse(NamedVertex[],Vector,Visitor) 
     * traverse} to take a possible limited recursion depth
     * into account.
     */
    private class ListEntry {

	public NamedVertex    vertex ;
	public int       depth  ;

	public ListEntry(NamedVertex aNode,int aDepth) {
	    vertex = aNode  ;
	    depth  = aDepth ;
	} // EndConstructor ListEntry

    } // EndInnerClass ListEntry

    /**
     * limits the recursion depth during the breath first traversal.
     * -1 is the default value and means that there is no limit.
     */
    private int _depthLimit = -1 ;

    public int traverse(NamedVertex[] startatVertices, 
			Vector   visited, 
			Visitor  visitor ) {
	NamedVertex    next;
	NamedVertex    adjacent;
	ListEntry    nextEntry ;
	Collection   outEdges;
	NamedEdge     edge;
	Iterator     iterator;

	// CHANGE FROM ORIGINAL IMPLEMENTATION ftaiani 2001-06-04

// 	if (_depthLimit==0) return OK ;

	// Put the starting *vertices* onto the list
	// Only the vertices which actually belong to the
	// graph is which the traversal is done are considered

// 	System.out.println("DFM: _depthLimit="+_depthLimit);

	for(int i = 0; i < startatVertices.length; i++)
	    if (this.dgraph.doesContainVertex(startatVertices[i]))
		this.list.add( new ListEntry(startatVertices[i],1) );

	while( !this.list.isEmpty() ){
	    // Get the next vertex in the list and add it to the visited
	    nextEntry = (ListEntry)this.list.remove(0);

	    next = nextEntry.vertex ;
	    visited.add( next );

	    if (   ( _depthLimit >= 0) && 
		   ( _depthLimit < nextEntry.depth) ) {
// 		System.out.println("DFM: Stopping on "+next ) ;
		continue ;
	    }

	    // Exit if the visitor tells us so
	    if( !visitor.visit( next ))
		return TERMINATEDBYVISITOR;


	    // Get all of its adjacent vertices, respecting the edge direction,
	    // and put them onto the list only if it has not been visited and
	    // it has not been listed

	    if (_isDirect) // Direct traversal
		outEdges = this.dgraph.getOutgoingEdges( next );
	    else          // Reverse traversal
		outEdges = this.dgraph.getIncomingEdges( next );

	    iterator = outEdges.iterator();
	    while( iterator.hasNext()) {
		edge = (NamedEdge) iterator.next();
		adjacent = (NamedVertex) edge.getOppositeVertex( next );
		if( !   visited.contains( adjacent ) && 
		    !this.list.contains( adjacent ) ) 
		    {
			if (adjacent.isPseudo())
			    this.list.add
				(0, new ListEntry(adjacent,nextEntry.depth) );
			else
			    this.list.add
				( new ListEntry(adjacent,nextEntry.depth+1) );
		    }
	    }
	} // EndWhile !this.list.isEmpty()

	return OK;
    }

    public Vector traverse( NamedVertex[] startatS ) {
	return this.traverse( startatS, new NullVisitor());
    }

    public Vector traverse( NamedVertex[] startatS, Visitor visitor ) {
	Vector  visited = new Vector();

	this.traverse( startatS, visited, visitor );
	return visited;
    }

} // EndClass BFMultiStartDirectedGraphTraversal
