/*
 * Copyright 2004 Francois Taiani <f.taiani@computer.org>
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * CallGraphBuilder.java
 * 
 * Time-stamp: <>
 * @author Francois Taiani   <f.taiani@computer.org>
 * $Id: CallGraphBuilder.java,v 1.3 2004/10/06 20:27:42 ftaiani Exp $
 * 
 */

package laas.friends.cosmopen.graphAnalysis  ;

import org.xml.sax.Attributes;
import laas.friends.cosmopen.utils.PropertyBag ;
import java.util.* ;

/** 
 * This class encapsulates some of the low level work involved in the
 * creation of a {@link CallGraph}. It can be used when creating a
 * graph on the fly or from a sequence of stack traces. It is used in
 * particular by {@link XMLTraceSet2CallGraphReader} and {@link
 * OnTheFlyCallGraphBuilder}.
 * @author Francois Taiani <f.taiani@computer.org>
 */
public class CallGraphBuilder {

    private String      usualThreadCreationPoint = null ;

    /**
     * Represents a virtual "top" vertex in the graph under
     * construction. It is not visible in this resulting graph but
     * makes the building process more convenient.
     */
    public NamedVertex topVertex ;

    private Map         mapOfPendingThreadCreationPointS = new TreeMap() ;
    private String      eventID   = null ;
    private NamedVertex lastCallAdded = null ;
    private NamedVertex lastLikelyThreadCreationCall = null ;
    private CallGraph   referenceGraph ;

    protected CallGraph receiver = null ;
    protected CallGraph context  = null ;

    public CallGraphBuilder(CallGraph someReceiver, CallGraph someContext) {
        receiver = someReceiver ;
        context  = someContext  ;

	if ( context !=null ) referenceGraph = (CallGraph)context   ;
	else                  referenceGraph = (CallGraph)receiver  ;

	topVertex      = ((CallGraph)referenceGraph).hiddenTopVertex ;

    } // EndConstructor CallGraphBuilder

    /**
     * indicates a method or function that is used to spawned new
     * threads, such as "::clone" on Linux, or
     * "java.lang.Thread::start" in Java. This helps the {@link
     * CallGraphBuilder} detect more accurately where a thread might
     * have been spawned.
     */
    public void setUsualThreadCreationPoint(String aThreadCreationPoint) {
        usualThreadCreationPoint = aThreadCreationPoint ;
    }

    public ChildrenRegistry getChildrenRegistryForVertex(NamedVertex aNVertex) {
        return referenceGraph.getChildrenRegistryForVertex(aNVertex) ;
    }

    /**
     * Stores information about a call occurence.
     */
    public class EventInfo {
        private String   callPoint = null ;
        public  String   threadID ;

	PropertyBag callAttributes = new PropertyBag();
	public EventInfo(Attributes someCallAttributes, 
                        String     currentStackThread) {
	    AbstractXMLReader.putSAXAttributesInPropertyBag
		    (someCallAttributes ,  callAttributes );
            threadID = currentStackThread ;
	} 
        public String getCallPointString() {
            if (callPoint==null)
                callPoint   = ChildrenRegistry.getNameOfCallPoint
                    ( this.callAttributes.getValueForProperty("entity"),
                      this.callAttributes.getValueForProperty("method")
                      ) ;
            return callPoint ;
        }
	public EventInfo(String entity, 
                        String method,
                        String thisRef, 
                        String currentStackThread) {

            threadID = currentStackThread ;
            callAttributes.setPropertyToValue("entity",entity );
            callAttributes.setPropertyToValue("method",method );
            callAttributes.setPropertyToValue("this"  ,thisRef);

	} 
    } // EndClass EventInfo

    public class ThreadCreationInfo {
	public NamedVertex creator ;
	public String      evendID ;
	public ThreadCreationInfo(NamedVertex aCreator ,String anEventID) {
	    creator   = aCreator  ;
	    evendID   = anEventID ;
	}
	public String toString() { return "Creation:"+evendID ; }
    } // EndClass ThreadCreationInfo


    private NamedVertex getMostLikelyCreationPoint() {

	NamedVertex result ;
	if (lastLikelyThreadCreationCall!=null) {
	    result = lastLikelyThreadCreationCall ;
	} else
	    result = lastCallAdded ;

	return result ;
    } // EndMethod getMostLikelyCreationPoint

    /** **************************************************** */
    public void registerThreadCreationPoint(String threadID) {
	mapOfPendingThreadCreationPointS.put
	    (  threadID,
	       new ThreadCreationInfo(getMostLikelyCreationPoint(),getNewEventID())
	       ) ;
    } // EndMethod registerThreadCreationPoint

    /** **************************************************** */
    public void registerThreadTerminationPoint(String threadID) {
        String terminationID = getNewEventID() ;

        try {
            receiver.addThreadTermination(threadID,terminationID);

            if (context!=null) 
                context.addThreadTermination(threadID,terminationID);
        } catch (Exception e) {
            e.printStackTrace() ;
        } // EndIf alreadyInTree

    } // EndMethod registerThreadTerminationPoint

    /** **************************************************** */
    private String getNewEventID() {
	return String.valueOf(++((CallGraph)receiver).greatestEventIDInGraph) ;
    }

    /** **************************************************** */
    public static String getThreadTag(String threadID) {

	if (threadID!=null) 
	    return "t"+threadID+"-" ;
	else
	    return "" ;

    } // EndMethod getThreadTag

    /** 
     * Creates new NamedVertex representing an invocation. It returns
     * the new vertex that was added.
     */
    public NamedVertex createNewEvent(EventInfo   newEventInfo, 
                                      NamedVertex parentOfNewEvent)
    {
        boolean alreadyInTree = (parentOfNewEvent!=topVertex) ;
        boolean thisIsAThreadStartPoint =false ;

	ThreadCreationInfo creationPoint = null ;
        
        // If we have just switched to a new thread, then this is a
	// thread starting point, and we have already registered a an
	// event for this point. Otherwise, this is a new
	// event. 

	if (!alreadyInTree) { 
	    // At top of the tree, so maybe a thread creation link must be created
	    thisIsAThreadStartPoint = 
		mapOfPendingThreadCreationPointS.containsKey(newEventInfo.threadID);
	} // EndElse at top of a tree

	if (thisIsAThreadStartPoint) {
	    // We're at the start point of a thread, so this is not a new event.
	    // we retreive the eventID from the creation point.
	    creationPoint = 
		(ThreadCreationInfo)
                mapOfPendingThreadCreationPointS.get(newEventInfo.threadID);

	    eventID = creationPoint.evendID ;
	} // EndIf thisIsAThreadStartPoint
	else
	    eventID = getNewEventID() ;

	NamedVertex newEventVertex = 
	    new NamedVertex(newEventInfo.getCallPointString()+"'"+
                            getThreadTag(newEventInfo.threadID)+eventID) ;

	newEventVertex.getPropBag().addBag(newEventInfo.callAttributes);
	newEventVertex.setPropertyToValue("eventOrder",eventID) ;
	if (newEventInfo.threadID!=null)
	    newEventVertex.setPropertyToValue("threadID" ,newEventInfo.threadID) ;

	receiver.add(newEventVertex) ;
        lastCallAdded = newEventVertex ;

	if (context!=null) context.add(newEventVertex) ;

	// Connecting the new event occurence to its parent.
	// If we are within the trace (alreadyInTree==true) , we connect it to the
	// previous parent.
	// If we are at a thread start point, we connect it to the event where the
	// thread was spawned.

	if (alreadyInTree)
	    try {
		receiver.addCallFromTo(parentOfNewEvent,newEventVertex) ;
		if (context!=null) 
		    context.addCallFromTo(parentOfNewEvent,newEventVertex) ;
	    } catch (Exception e) {
		e.printStackTrace() ;
	    } // EndIf alreadyInTree

	if (thisIsAThreadStartPoint) {
	    NamedVertex creator = creationPoint.creator ;

	    mapOfPendingThreadCreationPointS.remove(newEventInfo.threadID);
	    if (creator!=null)
		try {
		    ((CallGraph)receiver)   
			.addThreadCreationFromTo(creator,newEventVertex) ;
		    if (context!=null) 
			((CallGraph)context)
			    .addThreadCreationFromTo(creator,newEventVertex) ;
		} catch (Exception e) {
		    e.printStackTrace() ;
		} 
	} // EndIf thisIsAThreadStartPoint

	if (usualThreadCreationPoint!=null) {
	    if (ChildrenRegistry.getNameOfCallPoint(newEventVertex)
		.equals(usualThreadCreationPoint) )  {
			    
		lastLikelyThreadCreationCall = newEventVertex ;

	    } // newEventVertex is an usualThreadCreationPoint
	} // an indication of usualThreadCreationPoint was in the trace set header

	return newEventVertex ;

    } // EndMethod createNewEvent

} // EndClass CallGraphBuilder
