/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * ChildrenRegistry.java
 * 
 * Time-stamp: <2003-03-07 18:25:36 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: ChildrenRegistry.java,v 1.5 2003/10/13 12:18:19 ftaiani Exp $
 * 
 */
package laas.friends.cosmopen.graphAnalysis  ;

import java.util.* ;

/**
 * This class is used by {@link ChildrenRegistry} to
 * associate each "node" with the "youngest" of a child
 * for a particular thread. "Youngest" means the most
 * recently added child.<p>
 * This information is used in {@link XMLTraceSet2CallGraphReader}
 * to decide when a stack trace frame corresponds to a new
 * call or is only the repetition of a call that has already
 * been identified in a previous stack trace.
 */
class ThreadAndEventIDChildrenMap {
    private Map threadToChildrenBucket = new HashMap() ;

    public void addChild(NamedVertex child) {
	String threadOfNewChild = child.getValueForProperty("threadID") ;
	Map bucketForThread = (Map)threadToChildrenBucket.get(threadOfNewChild) ;
	if (bucketForThread==null) {
	    bucketForThread = new TreeMap() ;
	    threadToChildrenBucket.put(threadOfNewChild,bucketForThread) ;
	} // EndIf first time we have a new child with that thread
	bucketForThread.put(Integer.valueOf(child.getValueForProperty("eventOrder")), child) ;
    } // EndMethod addChild

    public void removeChild(NamedVertex child) {
	String threadOfNewChild = child.getValueForProperty("threadID") ;
	// threadOfNewChild==null is OK. It means no thread is
	// known for this event. Handled as if it was an extra thread.
	Map bucketForThread = (Map)threadToChildrenBucket.get(threadOfNewChild) ;
	if (bucketForThread==null) return ; // No child known with this thread ID
	bucketForThread.remove(Integer.valueOf(child.getValueForProperty("eventOrder"))) ;
	if (bucketForThread.isEmpty())
	    threadToChildrenBucket.remove(threadOfNewChild) ;
    } // EndMethod removeChild

    public NamedVertex getYoungestChildForThread(String threadID) {
	SortedMap bucketForThread = (SortedMap)threadToChildrenBucket.get(threadID) ;
	if (bucketForThread==null) return null ; // No child known with this thread ID
	return (NamedVertex)bucketForThread.get( bucketForThread.lastKey() ) ;	
    }
	
} // EndClass ThreadAndEventIDChildrenMap

/**
 * <code>ChildrenRegistry</code> keeps information about
 * the "child" events of a node in a {@link CallGraph}. This
 * information is not included in {@link NamedVertex} because
 * a {@link NamedVertex} can be in different {@link CallGraph graph},
 * and this kind of information strongly depends on the graph
 * it is in.<p>

 * This information is used to efficiently manage the set of call
 * trees contained in a {@link CallGraph}. One {@link
 * ChildrenRegistry} is associated with each vertex, and a global one
 * (called {@link CallGraph#topChildrenMap}) exists for the graph. A {@link
 * ChildrenRegistry} helps find among its children any child that
 * correspond to a given call point (a Class + a Method of that
 * class).  The global {@link ChildrenRegistry} of the graph does the
 * same for the vertices with not parents in the graph (the roots of
 * the trees of the graph).

 * @see CallGraph
 * @version $Name: release-1-3-rc1 $ $Revision: 1.5 $
 * @since Version 0.1.1 * 
 */
public class ChildrenRegistry {

    private Map       callPtToChildrenMap = new TreeMap() ;
    //    private SortedMap evntIDtoChildrenMap = new TreeMap() ;
    // Children are sorted according to their thread.
    private ThreadAndEventIDChildrenMap 
	threadToYoungestChild = new ThreadAndEventIDChildrenMap() ;

    public void addChild(NamedVertex child) {
	callPtToChildrenMap. put( getNameOfCallPoint(child)              , child ) ;
	threadToYoungestChild.addChild(child) ;
// 	evntIDtoChildrenMap. put( child.getValueForProperty("eventOrder"), child ) ;
    }

    public void removeChild(NamedVertex child) {
	callPtToChildrenMap.remove(getNameOfCallPoint(child)) ;
	threadToYoungestChild.removeChild(child) ;
// 	evntIDtoChildrenMap.remove(child.getValueForProperty("eventOrder")) ;
    }

    public boolean doesContainChildWithCallPoint(String nameOfCallPoint) {
	return callPtToChildrenMap.containsKey(nameOfCallPoint) ;
    }
    

    public NamedVertex getYoungestChildForThread(String threadID) {
	return threadToYoungestChild.getYoungestChildForThread(threadID) ;
    } // EndMethod getYoungestChildForThread

    /** ****************************************************
     * used to decide if a stack frame in a stack trace is to become
     * a new event or not. If the youngest "child" event that
     * is registered in the {@link ChildrenRegistry} has the same
     * "callpoint" (a call point is basically an entry point in the code),
     * then we don't consider the new frame to correspond to a new
     * invocation. Otherwise we do.
     * @see CallGraph#addCallFromTo(NamedVertex caller,NamedVertex callee)
     * @see CallGraph#addThreadCreationFromTo(NamedVertex caller,NamedVertex callee)
     */
    public boolean youngestChildForThreadHasSameCallPoint( String threadID, 
							   String nameOfCallPoint) 
    {
//   	System.out.println("*** IN HASYOUNGESTCHILDSAMECALLPOINT:");
//   	System.out.println(" -> nameOfCallPoint : "+nameOfCallPoint) ;
	NamedVertex youngestForThread 
	    = threadToYoungestChild.getYoungestChildForThread(threadID) ;
	if (youngestForThread==null) {
	    // No children, so no youngest child...
//  	    System.out.println(" -> No children, so no youngest child...");
	    return false ;
	}
	String callPointOfYoungest 
	    = getNameOfCallPoint( youngestForThread ) ;
//  	System.out.println(" -> callPointOfYoungest : "+callPointOfYoungest) ;
	return callPointOfYoungest.equals(nameOfCallPoint) ;
	
    } // Method hasYoungestChildSameCallPoint

    public NamedVertex getChildForCallPoint(String nameOfCallPoint) {
	return (NamedVertex)callPtToChildrenMap.get(nameOfCallPoint) ;
    }

    public static String getNameOfCallPoint(NamedVertex toBeNamed) {
	return   getNameOfCallPoint
	    ( toBeNamed.getValueForProperty("entity"),
	      toBeNamed.getValueForProperty("method") ) ;
    } // EndMethod getNameOfCallPoint

    public static String getNameOfCallPoint(String entity, String method ) {

	return entity + "::" + method ; 

    } // EndMethod getNameOfCallPoint

    public String toString() {
	return "ChildrenRegistry: "+ callPtToChildrenMap ; 
    } // EndMethod toString

    public int size() {
	return callPtToChildrenMap.size() ;
    } // EndMethod size

} // EndClass ChildrenRegistry
