/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * DGWriterAdapter.java
 * 
 * Time-stamp: <2002-08-16 22:15:54 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: DGWriterAdapter.java,v 1.4 2007/09/18 18:23:42 ftaiani_normal Exp $
 * 
 */
package laas.friends.cosmopen.graphAnalysis  ;


/**
 * This class is the abstract root class of all
 * classes implementing the output of {@link DirectedGraphWithDict }
 * instances to a stream in a given format.
 * @see DirectedGraphWithDict2DotOStream
 * @see XMLGraphWriter
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.4 $
 * @since Version 0.0.4
 */
public abstract class DGWriterAdapter implements DirectedGraphWriter {

    /**
     * The stream the writer writes to. This variable
     * should be used by subclass to directed interact
     * with the stream.
     */
    protected java.io.PrintStream _theStreamIWriteTo ;

    public void setOutstream(java.io.OutputStream anOStream) {
	_theStreamIWriteTo = new java.io.PrintStream(anOStream) ;
    } 

    public DGWriterAdapter() {} 

    public DGWriterAdapter(java.io.OutputStream anOStream) 
    {
	_theStreamIWriteTo = new java.io.PrintStream(anOStream) ;
    } // EndConstructor DGWriterAdapter

    private DirectedGraphWithDict _theGraphIWriteFrom = null ;

    protected DirectedGraphWithDict theGraphBeingWrittenOut() {
	return _theGraphIWriteFrom ;
    }

    /**
     * This is the main methods of the class. It goes through the passed
     * {@link  DirectedGraphWithDict} and produces an output on the
     * {@link java.io.OutputStream OutputStream } passed in the
     * {@link #DGWriterAdapter(java.io.OutputStream) constructor }.
     * The output is based on the smaller methods writeSomething
     * of the class, which have to be overwritten for concrete
     * implementations.
     * @see #writeHeader(DirectedGraphWithDict)
     * @see #writeClosing()
     * @see #writeVertexRendering(NamedVertex)
     * @see #writeVertexClosing()
     * @see #writeEdgeRendering(NamedEdge)
     */
    public void writeGraph(DirectedGraphWithDict aGraphToWrite) {

	_theGraphIWriteFrom = aGraphToWrite ;
	writeHeader(aGraphToWrite) ;

	writeGraphSwitchVertexSwitchLink(aGraphToWrite,true,true);

	writeClosing() ;
       	_theGraphIWriteFrom = null ;

    } // EndMethod writeGraph

    public void writeGraphVertexSeparate(DirectedGraphWithDict aGraphToWrite) {

	_theGraphIWriteFrom = aGraphToWrite ;
	writeHeader(aGraphToWrite) ;

	// Depending on whether the order of edges should be
	// proeminant for the representation of the graph we draw the
	// vertices first or last

	if (!aGraphToWrite.isEdgeOrderProeminant())
	    writeGraphSwitchVertexSwitchLink(aGraphToWrite,true,false);

	java.util.Iterator iterThroughNamedEdge
		    = aGraphToWrite.getEdgeIterator() ;
	writeEdgesUsingIterator (iterThroughNamedEdge);
	
	if (aGraphToWrite.isEdgeOrderProeminant())
	    writeGraphSwitchVertexSwitchLink(aGraphToWrite,true,false);

	writeClosing() ;
       	_theGraphIWriteFrom = null ;

    } // EndMethod writeGraph

    public void writeGraphSwitchVertexSwitchLink(DirectedGraphWithDict aGraphToWrite,
						 boolean switchVertex,
						 boolean switchLink ) 
    {
	// *** LOOP THROUGH ALL CLASSDATA OF THE GRAPH

	java.util.Iterator iterThroughNamedVertex 
	    = aGraphToWrite.getVerticesIterator() ;

	while (iterThroughNamedVertex.hasNext() ) {

	    NamedVertex currentNamedVertex 
		= (NamedVertex)iterThroughNamedVertex.next() ;

	    if (switchVertex) writeVertexRendering(currentNamedVertex) ;
	    if (switchLink) {
		java.util.Iterator iterThroughNamedEdge
		    = aGraphToWrite.getOutgoingEdges(currentNamedVertex).iterator() ;
		writeEdgesUsingIterator (iterThroughNamedEdge);
	    }
	    if (switchVertex) writeVertexClosing() ;

	} // EndWhile Loop through the vertices of the graph

    } // EndMethod writeGraphToStream

    private void writeEdgesUsingIterator(java.util.Iterator iterThroughNamedEdge)
    {
	// *** LOOP THROUGH ALL LINKDATA OF THE CURRENT CLASSDATA
	
	while (iterThroughNamedEdge.hasNext() ) {
	    NamedEdge currentNamedEdge 
		= (NamedEdge)iterThroughNamedEdge.next() ;
	    // Writes the edge in a dot fashion
	    //	    System.out.println("Rending Edge "+currentNamedEdge);
	    writeEdgeRendering(currentNamedEdge) ;
	    
	} // EndWhile Loop through the link of the current NamedVertex

    } // EndMethod writeEdgesForVertex

    /**
     * writes the header of the output file to the stream.
     * Default implementation does nothing.
     */
    protected void writeHeader(DirectedGraphWithDict graphToBeRendered) {}

    /**
     * write the closing of the output file to the stream.
     * Default implementation does nothing.
     */
    protected void writeClosing() {}

    /**
     * renders the properties of the passed {@link NamedVertex } in the
     * output format.
     */
    protected abstract void writeVertexRendering(NamedVertex classToBeRendered) ;

    /**
     * write the closing of a class to the stream. For instance
     * in XML, should write the closing tag "</class>" to the stream.
     * Default implementation does nothing.
     */
    protected void writeVertexClosing() {}

    /**
     * renders the properties of the passed {@link NamedEdge }  in the
     * output format.
     */
    protected abstract void writeEdgeRendering(NamedEdge linkToBeRendered) ;

} // EndClass DGWriterAdapter
