/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * DirectedGraphWithDict.java
 * 
 * Time-stamp: <2003-03-07 22:18:02 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: DirectedGraphWithDict.java,v 1.7 2007/09/18 18:23:42 ftaiani_normal Exp $
 * 
 */
package laas.friends.cosmopen.graphAnalysis  ;

import salvo.jesus.graph.Vertex ;
import laas.friends.cosmopen.utils.PropertyBag ;

/**
 * This interface adds dictionnaries facilities for
 * {@link NamedVertex } vertices to a
 * a {@link salvo.jesus.graph.DirectedGraph}.<P>
 * <B>Note:</B><BR>
 * A {@link DirectedGraphWithDict } is assumed
 * to contained only {@link salvo.jesus.graph.Vertex vertices}
 * of the type {@link NamedVertex }. A {@link DirectedGraphWithDict }
 * can't be substituted to a {@link salvo.jesus.graph.DirectedGraph }
 * in a method which does not respect that constraint.
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.7 $
 * @since Version 0.0.2
 */
public interface DirectedGraphWithDict 
extends salvo.jesus.graph.DirectedGraph, Cloneable
{
    /**
     * This exception is thrown by 
     * {@link #getNodeSWithRegExp(String limitedWildcardExp ) }
     * when the passed limitedWildcardExp is ill formed.
     * @see #getNodeSWithRegExp(String limitedWildcardExp )
     */
    public class NotALimitedWilcardExpression 
	extends java.lang.IllegalArgumentException
    {
	public NotALimitedWilcardExpression(String message) {
	    super(message) ;
	} // EndConstructor NotALimitedWilcardExpression
    }

    public boolean doesContainVertex(Vertex candidate) ;

    /**
     * Adds a collection a collection of {@link NamedVertex NamedVertexS } 
     * to the graph.
     */
    public void addNamedVertexS(java.util.Collection classDataS) ;

    /**
     * Adds the nodes and edges of another {@link DirectedGraphWithDict }
     * to the current graph. The dictionnary of the current graph is 
     * updated.
     */
    public void addGraph(DirectedGraphWithDict aGraphToAdd) 
	throws java.lang.Exception ;

    /**
     * Add a {@link NamedVertex} instances to the {@link DirectedGraphWithDict}.
     * The {@link NamedVertex} is added to the graph and to the dictionnary
     * using its name. Names must be unique within a graph. In case a different
     * {@link NamedVertex} is already present in the graph with the same name
     * a {@link java.lang.RuntimeException} is raised. On the contrary
     * if the passed {@link NamedVertex newNamedVertex} is already in
     * the {@link DirectedGraphWithDict}, no exception is thrown.
     * <p><b>NOTE:</b>A {@link DirectedGraphWithDict} can only contains
     * {@link NamedVertex} instances. If this method is called with
     * object which are not instances of {@link NamedVertex} a
     * run time exception is thrown.
     */
    public void add(salvo.jesus.graph.Vertex newNamedVertex) ;

    /**
     * Searches for a {@link NamedVertex} with the passed name
     * in the {@link DirectedGraphWithDict}. If none is found
     * null is returned.
     * @see #provideNamedVertexWithName(String aName)
     */
    public NamedVertex getNodeWithName(String aName) ; 

    /**
     * Searches is there is already a {@link NamedVertex} with the passed name
     * in the {@link DirectedGraphWithDict}. If yes this {@link NamedVertex}
     * is returned. If no, a new {@link NamedVertex} is created
     * and added to the {@link DirectedGraphWithDict}
     * with the corresponding name.
     * @see #getNodeWithName(String aName)
     */
    public NamedVertex provideNamedVertexWithName(String aName) ; 

   /**
     * removes the edges from the graph which are 
     * provided by an iterator passed as parameter. This iterator
     * should contain only {@link NamedEdge }
     */
    public void removeEdgesWithIterator(java.util.Iterator iter) 
    	throws Exception ;

   /**
     * removes the vertices from the graph which are 
     * provided by an iterator passed as parameter. This iterator
     * should contain only {@link NamedVertex }.
     * @return The number of removed vertices.
     */
    public int removeVertexWithIterator(java.util.Iterator iter)
        throws Exception ;

    /**
     * returns an iterator which goes through all the edges of the
     * graph.
     */
    public java.util.Iterator getEdgeIterator() ;

    /**
     * returns an iterator which goes through all the edges of the
     * graph of a certain kind. If the <code>kind</code> parameter
     * is null, returns all edges.
     */
    public java.util.Iterator getEdgeIteratorKind(String kind) ;

    /**
     * returns true if the passed vertex has either no parents or no child
     * in the graph.
     */
    public boolean vertexHasNoParentsOrNoChilds(NamedVertex candidate) ;

    /**
     * removes all vertices adjacent to <code>needsRoom</code>
     * which are only connected to <code>needsRoom</code>.
     * @param needsRoom a {@link NamedVertex } from which to take
     *                  'weak' neighbors.
     */
    public void cleanAroundVertex(NamedVertex needsRoom)
	throws java.lang.Exception ;
    /**
     * This methods extracts the subgraph obtained when traversing
     * the graph along the directed edges in the reverse direction
     * from the sink vertex to the source from a given starting vertex.
     */
    public DirectedGraphWithDict 
	extractBackwardClosure(NamedVertex aStartingPoint) ;

    /**
     * This methods extracts the subgraph obtained when traversing
     * the graph along the directed edges (in the normal direction
     * source to sink) from a given starting vertex.
     */
    public DirectedGraphWithDict 
	extractForwardClosure(NamedVertex aStartingPoint) ;

    /**
     * Does the same as 
     * {@link #extractBackwardClosure(NamedVertex aStartingPoint) 
     * extractBackwardClosure } but starting from a subgraph.
     * In case some nodes of the <code>aStartingGraph</code>
     * are no contained in the called graph, no exception is thrown.
     */
    public DirectedGraphWithDict 
	extractBackwardClosureFromGraph(DirectedGraphWithDict aStartingGraph) ;

    /**
     * Does the same as 
     * {@link #extractForwardClosure(NamedVertex aStartingPoint) 
     * extractBackwardClosure } but starting from a subgraph.
     * In case some nodes of the <code>aStartingGraph</code>
     * are no contained in the called graph, no exception is thrown.
     */
    public DirectedGraphWithDict 
	extractForwardClosureFromGraph(DirectedGraphWithDict aStartingGraph) ;

    /**
     * Similar to {@link #extractForwardClosureFromGraph(DirectedGraphWithDict)
     * extractForwardClosureFromGraph}, but from a cloud of nodes.
     */
    public DirectedGraphWithDict
	extractForwardClosureFromArray(NamedVertex[] someStartingPointS) ;

    /**
     * Similar to {@link #extractBackwardClosureFromGraph(DirectedGraphWithDict)
     * extractBackwardClosureFromGraph}, but from a cloud of nodes.
     */
    public DirectedGraphWithDict
	extractBackwardClosureFromArray(NamedVertex[] someStartingPointS) ;

    /**
     * computes the backward set of a seed graph within the context
     * of the called {@link DirectedGraphWithDict}, within a given
     * step limit. The backward closure whould correspond to an
     * infinite step limit.
     */
    public DirectedGraphWithDict 
	extractFixBackwardFromGraph(DirectedGraphWithDict aStartingGraph,
				    int                   depthLimit ) ;
    /**
     * computes the forward set of a seed graph within the context
     * of the called {@link DirectedGraphWithDict}, within a given
     * step limit. The forward closure whould correspond to an
     * infinite step limit.
     */
    public DirectedGraphWithDict 
	extractFixForwardFromGraph(DirectedGraphWithDict aStartingGraph,
				    int                  depthLimit ) ;
    /**
     * gets the {@link NamedVertex } nodes with a name matching the
     * limited wildcard expression. Limited means the expression may
     * contain a star (*) either at its end or at its beginning.
     * Other configuration are vorbidden. For example: "*abDC" means
     * all names ending with "abDC". "*xxx*" means all names containing
     * "xxx".
     * This method should throw a {@link NotALimitedWilcardExpression }
     * each time the passed expression is ill formed (for instance "**",
     * or "*X*Y*Z*").
     * @see NotALimitedWilcardExpression
     */
    public java.util.SortedSet
	getNodeSWithRegExp(String limitedWildcardExp) ;

    /**
     * Returns the link of the graph with the given source, sink
     * and name, if it exists. If there is no such link, null is returned.
     */
    public NamedEdge getLinkWithSourceSinkName(NamedVertex   source,
					      NamedVertex   sink  ,
					      String      name) ;

    /**
     * adds edges to the <code>this</code> graph using the context graph.
     * For all vertices which are present in the called object as
     * weel as in the passed context graph, if an edge exists between 
     * those vertices in the context but not in the called
     * object, that edge is added to the called object if it. This
     * can also be seen as a kind of envelop computation.
     */
    public void 
	doEdgeCompletionInContext(DirectedGraphWithDict Context) 
	throws java.lang.Exception ;

    public Object clone() ;

    /** *******************************************
     * As {@link Object#equals(Object)} excpet that in case
     * the graphs are no equal, some diagnostic is printed.
     */
    public boolean equalsDebug(Object obj) ;

    public DirectedGraphWithDict getEmptyGraphOfSameType ( ) ;
    /**
     * returns a {@link DirectedGraphWriter} associated with the passed
     * {@link java.io.OutputStream}, that formats the graph according
     * to its type.
     */
    public DirectedGraphWithDict2DotOStream getFormattingWriter( ) ;
    /**
     * As {@link #getFormattingWriter()}, but
     * the return {@link DirectedGraphWriter} does format the graph at all. (raw mode.)
     */
    public DirectedGraphWithDict2DotOStream getRawWriter( ) ;

    /**
     * returns the {@link PropertyBag} of the graph.
     */
    public PropertyBag getPropBag() ;

    public int getOutDegree (Vertex candidate) ;
    public int getInDegree  (Vertex candidate) ;
    public int getDegree    (Vertex candidate) ;

    /**
     * Is used when printing out the graph to transform the vertex
     * names that are use internally into a human friendly
     * representation ("desentangling"). 'removeNamespaceFor' tells is
     * the name space part of entity names is to be removed. For
     * graphs whose internal representation is already human readable,
     * this method simply returns the internal name of the passed
     * vertex (after possibly removing the namespace part when requested).
     */
    public String getHumanFriendlyName(NamedVertex toBeNamed,
				       boolean     removeNamespace) ;

    /**
     * Is called by 
     * {@link DirectedGraphWithDict2DotOStream.writeGraphInDotFormat(DirectedGraphWithDict)} to
     * allow the current graph to generate any needed resources for
     * the visualisation (typically temporaray IDs)
     */
    public void prepareForVisualisation() ;

    /**
     * Releases any resources created by {@link prepareForVisualisation()}
     */
    public void endOfVisualisation();

    /**
     * When writing the graph out should the order of the edges
     * decides the order in which the graph is drawn rather than the
     * order of the vertices.
     */
    public boolean isEdgeOrderProeminant();

} // EndInterface DirectedGraphWithDict
