/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * XMLGraphBuilder.java
 * 
 * Time-stamp: <2003-03-06 16:15:13 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: XMLGraphBuilder.java,v 1.1 2004/09/09 15:34:29 ftaiani Exp $
 * 
 */
package laas.friends.cosmopen.graphAnalysis  ;

import salvo.jesus.graph.DirectedGraphImpl ;
// import salvo.jesus.graph.DirectedGraph ;

import java.io.InputStream ;
import laas.friends.cosmopen.utils.PropertyBag ;

/**
 * This class helps to constuct a {@link DirectedGraphWithDictImpl }
 * incrementally by adding {@link NamedVertex } and {@link NamedEdge } objects
 * to the graph from SAX XML parse informations. This class does "basic work"
 * which needs to be coordinated by a more abstract "constructions" driver.
 * Such a driver is implemented in {@link XMLGraphReader }.
 * @see XMLGraphReader
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.1 $
 * @since Version 0.0.1
 */
public class XMLGraphBuilder {
  
    private DirectedGraphWithDict _graphUnderWork ;
    private DirectedGraphWithDict _contextGraph   ;
    /** 
     * is equals to _contextGraph if there is an explicit context graph,
     * otherwise (i.e. _contextGraph==null ), refers directly to
     * _graphUnderWork
     */
    private DirectedGraphWithDict _referenceGraph ;
    private boolean               _iAmInANamedVertexConstruction = false ;
    private NamedVertex           _vertexUnderWork ;
    private boolean               _vertexUnderWorkNewToContext ; 

    public XMLGraphBuilder() {
	_contextGraph = null ;
    } // EndConstructor

    /**
     * Provides a context graph for the construction. With a context
     * graph, new {@link NamedVertex} which are read from the xml file
     * are first searched in the context graph. If the context graph
     * contains a {@link NamedVertex} with the same name, this {@link NamedVertex }
     * is used. If not, a new one is instanciated, and added to the context
     * graph as well as to the result graph.
     */
    public XMLGraphBuilder(DirectedGraphWithDict contextGraph) {
	_contextGraph   = contextGraph ;
	_referenceGraph = contextGraph ;
    } // EndConstructor

    /**
     * can be overwritten to change the graph type used by the builder
     */
    public DirectedGraphWithDict createNewGraph() {
	return new DirectedGraphWithDictImpl() ;
    }

    /**
     * starts the construction of a directed graph in the 
     * {@link XMLGraphBuilder }.
     * @see #stopConstructionAndReleaseGraph()
     */
    public void startGraphConstruction() {
	_graphUnderWork = createNewGraph() ;
	if (_referenceGraph==null) _referenceGraph = _graphUnderWork ;
    } // End startGraphConstruction

    /**
     * Same as {@link #startGraphConstruction()}, but explicitly
     * provide a graph, which may already contain nodes and edges,
     * for further construction through the builder.
     */
    public void startGraphConstruction
	(DirectedGraphWithDict graphUnderWork) 
    {
	if (graphUnderWork==null) {
	    startGraphConstruction();
	    return ;
	}
	_graphUnderWork = graphUnderWork ;
	if (_referenceGraph==null) _referenceGraph = _graphUnderWork ;
    } // EndMethod startGraphConstruction

    public void startGraphConstruction
	(DirectedGraphWithDict graphUnderWork,org.xml.sax.Attributes someAttributes) 
    {
	startGraphConstruction(graphUnderWork) ;
	AbstractXMLReader.putSAXAttributesInPropertyBag
	    (someAttributes,_graphUnderWork.getPropBag());
    } // End startGraphConstruction

    /**
     * during construction starts adding a {@link NamedVertex } Vertex in
     * the graph.
     */
    public void startAddingNamedVertexFromXML
	( org.xml.sax.Attributes someAttributes ) 
    {
// 	System.out.println("IN startAddingNamedVertexFromXML") ;
// 	System.out.println("  -> "+someAttributes) ;
	if (_iAmInANamedVertexConstruction) 
	    throw new RuntimeException
		("Already in a NamedVertex Building session.");
	_iAmInANamedVertexConstruction = true ;

	// Tries to retrieve a NamedVertex vertex with the passed name
	// In case none is found, a new NamedVertex is created and added
	// to the graph.

	String aClassName = someAttributes.getValue("name");
	ProvideCDResult resultOfProvide
	    = this.provideNamedVertexWithName( aClassName ) ;

	_vertexUnderWork             = resultOfProvide.returnedNamedVertex ;
	_vertexUnderWorkNewToContext = resultOfProvide.isNewToContext ;

	AbstractXMLReader.putSAXAttributesInPropertyBag
	    ( someAttributes,
	      _vertexUnderWork.getPropBag() );

	onNewPropertiesInVertex(_vertexUnderWork) ;

    } // End startAddingNamedVertex

    /**
     * Used for the results of {@link #provideNamedVertexWithName(String) }
     */
    private class ProvideCDResult {
	NamedVertex returnedNamedVertex = null ;
	boolean   isNewToContext    = false ;
    } // EndInnerClass ProvideCDResult

    private ProvideCDResult provideNamedVertexWithName
	(  String    aClassName ) 
    {
// 	System.out.println("provideNamedVertexWithName") ;
	ProvideCDResult result = new ProvideCDResult() ;
	if (_contextGraph==null) {
// 	    System.out.println(" -> _contextGraph==null") ;
	    result.returnedNamedVertex 
		= _graphUnderWork.provideNamedVertexWithName(aClassName) ;
	}
	else {

	    result.isNewToContext 
		= ( _contextGraph.getNodeWithName(aClassName)==null) ;
	    result.returnedNamedVertex 
		=   _contextGraph.provideNamedVertexWithName(aClassName) ;
	    _graphUnderWork.add(result.returnedNamedVertex) ;

	} // EndElse _contextGraph!=null

	return result ;

    } // EndMethod provideNamedVertexWithName

    protected String   previousChainID = null ;
    protected NamedEdge previousLink = null ;

    /**
     * During the addition of a {@link NamedVertex} Vertex, add a link from
     * this {@link NamedVertex} to another {@link NamedVertex } according to
     * the content of the {@link org.xml.sax.Attributes someAttributes }
     * parameter.
     */
    public void addLinkFromXML
	( org.xml.sax.Attributes someAttributes ) {
// 	System.out.println("IN addLinkFromXML") ;
// 	System.out.println("  -> "+someAttributes) ;
	if (!_iAmInANamedVertexConstruction) 
	    throw new RuntimeException
		("A link can only be added while adding a NamedVertex item.");

	String targetName = someAttributes.getValue("target");
// 	System.out.println("  -> "+targetName) ;

	// Links have to be identified with a name which is unique
	// with respect to the pair (source,sink). If no attribute
	// called "name" is present in the link properties
	// we use the name of the target node.
	// This means that for usage links
	// (i.e. plain attributes) we use the attribute name.
	// For inheritance, the superclass name. This is useful
	// to merge XML files wich are partically overlapping.
	// Otherwise we would get duplicate links for the same
	// real link.

	String aLinkName ;
	aLinkName  = someAttributes.getValue("name");
	if (aLinkName==null)
	    aLinkName  = someAttributes.getValue("target");
	if (aLinkName==null)
	    aLinkName  = "";
// 	System.out.println("  -> "+aLinkName) ;

	ProvideCDResult resultOfProvide
	    = this.provideNamedVertexWithName( targetName ) ;
// 	System.out.println("  -> resultOfProvide: "+resultOfProvide) ;

	NamedVertex targetNamedVertex    = resultOfProvide.returnedNamedVertex ;
	boolean     targetNewToContext   = resultOfProvide.isNewToContext    ;

	NamedEdge newLink = null ;

	newLink = new NamedEdge(_vertexUnderWork,targetNamedVertex,aLinkName) ;
	AbstractXMLReader.putSAXAttributesInPropertyBag( someAttributes,
							 newLink.getPropBag() ) ;
	boolean linkIsAlreadyInContext = true ;
	NamedEdge foundLink  = null ;

	if ( _contextGraph!=null) {
	    linkIsAlreadyInContext = false ;
	    if ( !_vertexUnderWorkNewToContext &&
		 !targetNewToContext )

		{ /* There is a context graph, and source and
		     sink of the new link are known to it. We
		     go through all the outgoing edges of the
		     source in the graph under work and
		     check for equality with the new link. */
		     foundLink = 
			getTwinLinkWithinGraph
			   ( newLink, _contextGraph ) ;

		} /* EndIf There is a context graph, and source and
		     sink of the new link are known to it */
	} // EndIf there is a context graph

	else // The new graph is its own context.
	    {
// 		System.out.println("  -> The new graph is its own context") ;
		foundLink = 
		    getTwinLinkWithinGraph
		    ( newLink, _graphUnderWork  ) ;
	    }

	// There is already a sibling link in either the context or the
	// Graph under construction it self. The 2 links attributes may be
	// different have to be merged using addAttributesToLinkFromLink.

	if (foundLink!=null) {

	    // A SIBLING WAS FOUND
// 	    System.out.println("  -> A SIBLING WAS FOUND") ;
		
	    foundLink.addAttributesFromLink(newLink) ;
	    newLink = foundLink ;
	    linkIsAlreadyInContext = true ;

	} // A sibling was found.

	PropertyBag bagOfNewLink = newLink.getPropBag() ;
	String chainID = bagOfNewLink.getValueForProperty("chainID") ;
	bagOfNewLink.setPropertyToValue("chainID",null); // Removes "chainID"

 	if (chainID!=null) {

	    // There is a chainID attribute, which means the link
	    // is part of a call chain. We have to compute the
	    // order of the call on the sink (i.e. the callee),
	    // which we call the "incomingOrder"

	    if (!bagOfNewLink.hasProperty("incomingOrder")) {
		int order = _referenceGraph.getIncomingEdges( newLink.getSink() ).size()+1 ;
		bagOfNewLink.setPropertyToValue
		    ( "incomingOrder",
		      String.valueOf(Character.forDigit(9+order,Character.MAX_RADIX))
 		  ) ; // End bagOfNewLink.setPropertyToValue

	    } // Not yet an incomingOrder property

	    if (!chainID.equals(previousChainID)) {
		addChainingFromPortToLink("",previousLink) ;
	    }
	    else {
		NamedVertex jointClass = (NamedVertex)newLink.getSink() ;
		if (jointClass!=(NamedVertex)previousLink.getSource())
			throw new RuntimeException
			    ("The XML input contains chainID attributes, but "+
			     "the resulting link chains are broken! : \n"+
			     jointClass+" != "+(NamedVertex)previousLink.getSource()+
			     " chainID = " + chainID ) ;
		String incomingPort = bagOfNewLink.getValueForProperty("incomingOrder") ;

		addChainingFromPortToLink(incomingPort,previousLink) ;
	    } // EnfIf we are still in the same call chain.
	    
 	} //  End The newLink has an chainID property

	try {
// 	    System.out.println("  -> _graphUnderWork.addEdge( newLink ) " );
	    _graphUnderWork.addEdge( newLink ) ;
	    if (!linkIsAlreadyInContext)
		_contextGraph.addEdge( newLink ) ;
	} catch (Exception e) {
	    e.printStackTrace() ;
	} // EndCatch

	if (chainID!=null) { 
	    previousChainID = chainID ;
	    previousLink    = getTwinLinkWithinGraph( newLink, _graphUnderWork ) ;
	}
    } // End addLinkFromXML

    /**
     * Is used in {@link #addLinkFromXML(org.xml.sax.Attributes)}
     * to label the edges of caller / callee types with information 
     * on the order and nesting of calls. Incoming calls are labeled with a
     * port label ("a","b",...) and outgoing calls are labeled with
     * a sequence of chainings of the form <port-name>:<sequence-number>.
     * A chaining exists between 2 calls when they appear in the same backtrace.
     * The port-name of a chaining is the port label of the incoming call
     * of this chaining, the sequence number corresponds to the order
     * in which the chaining appear in the XML input file (usually
     * produced from a GDB trace script).
     */
    private void addChainingFromPortToLink(String port,NamedEdge link) {
	if (link==null) return ;
	if (link.isAlreadyConnectedToIncomingPort(port) ) return ;
	link.connectToIncomingPort(port) ;
	NamedVertex  jointClass = (NamedVertex)link.getSource() ;
	PropertyBag bagOfLink = link.getPropBag() ;
	jointClass.countOfLinkChainings++ ;
	String callOrder = bagOfLink.getValueForProperty("callOrder");
	if (callOrder==null) 
	    callOrder="" ;
	else
	    callOrder= callOrder + "," ;
	if (port.length()>0)
	    callOrder = callOrder + port +"." ;
	callOrder = callOrder +jointClass.countOfLinkChainings ;
	previousLink.getPropBag().setPropertyToValue("callOrder",callOrder) ;

    } // EndMethd addChainingFromPortToLink

    /**
     * Looks for a link with the same name, source and sink as another
     * link (a "twin" link).
     * @param original the link whose twin we are looking for.
     * @param aGraph   the graph in which we do the search.
     * @return the found Link if any. Otherwise null.
     */
    private NamedEdge getTwinLinkWithinGraph
	(NamedEdge              original,
	 DirectedGraphWithDict  aGraph) 
    {
	return aGraph.getLinkWithSourceSinkName( (NamedVertex)original.getSource(),
						 (NamedVertex)original.getSink  (),
						 original.getName  () ) ;
    } // EndMethod getSiblingLinkFromGraph

    /**
     * merges the attributes of 2 link.
     */
    private void addAttributesToLinkFromLink(NamedEdge link1, NamedEdge link2) {

	PropertyBag propertieS1 = link1.getPropBag() ;
	PropertyBag propertieS2 = link2.getPropBag() ;

	java.util.Iterator iter = propertieS2.iterator() ;

	while (iter.hasNext()) {
	    String currentProperty = (String)iter.next() ;

		propertieS1.changePropertyToValue
		    (currentProperty,
		     propertieS2.getValueForProperty(currentProperty) ) ;

	} // End While

    } // EndMethod addAttributesToLinkFromLink

    /**
     * does nothing by default. Can be overwritten.
     */
    protected void onNewPropertiesInVertex(NamedVertex luckyVertex) {} ;

    protected DirectedGraphWithDict getGraphUnderWork() {
	return _graphUnderWork ;
    } 

    /**
     * must be call each time a {@link NamedVertex } has been added
     * to the graph, symetrically to {@link 
     * #startAddingNamedVertexFromXML(org.xml.sax.Attributes) }.
     */
    public void endAddingNamedVertexFromXML() {
	if (!_iAmInANamedVertexConstruction) 
	    throw new RuntimeException
		("No matching call to startAddingNamedVertexFromXML()");
	_iAmInANamedVertexConstruction = false ;
    } // End endAddingNamedVertexFromXML

    /**
     * returns the graph consctructed with the addSomethin methods.
     * On return the Builder looses any reference on the graph
     * it has just constructed, so that multiple calls on that
     * method without ones on  {@link #startGraphConstruction() }
     * will return null values.
     * @see #startGraphConstruction()
     */
    public DirectedGraphWithDict stopConstructionAndReleaseGraph() {
	if (_iAmInANamedVertexConstruction) 
	    throw new RuntimeException
		("Trying to stop construction while still"+
		 "in a NamedVertex Building session.");
	// See addLinkFromXML to understand why it has to be there.
	addChainingFromPortToLink("",previousLink) ;
	DirectedGraphWithDict result = _graphUnderWork ;
	_graphUnderWork = null ;
	return result;
    } // End stopConstructionAndReleaseGraph

} // EndClass XMLGraphBuilder



