/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * TestXMLGraphReader.java
 * 
 * Time-stamp: <02/08/16 11:50:40 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: TestXMLGraphReader.java,v 1.2 2004/10/27 17:16:53 ftaiani Exp $
 * 
 */
package laas.friends.cosmopen.graphAnalysis.test  ;

import java.io.ByteArrayInputStream    ;

import laas.friends.cosmopen.graphAnalysis.* ;
import laas.friends.cosmopen.utils.PropertyBag ;

/**
 * Tests the class {@link XMLGraphReader }.
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.2 $
 * @since Version 0.0.1
 */
public class TestXMLGraphReader 
    extends BagBasedTest {

    private AbstractXMLReader         readerTestee ;
    private DirectedGraphWithDict     result ;
    private String                    inputString ;

    public TestXMLGraphReader(String name) {
	super(name) ;
    } // EndConstructor TestXMLGraphReader

    /**
     * Can be overwritten to test subclasses of AbstractXMLReader
     * that should bevahe as {@link XMLGraphReader}
     * within its definition domain.
     */
    protected AbstractXMLReader createGraphReaderTestee() {
	return new XMLGraphReader() ;
    }
    
    /**
     * Constructing from an empty document should produced 
     * an empty graph.
     */
    public void testConstructingFromEmpty() throws Exception
    {
	readerTestee = createGraphReaderTestee() ;
 	result = readerTestee.readGraphFromXML
	    (new ByteArrayInputStream
	     ( new String
	       ("<project></project>").getBytes()) 
	     ) ;
	assertTrue(result.getVerticesCount()==0) ;

    } // EndMethod testConstructingFromEmpty
    
    /**
     * Tries to build a graph containing only one class with no links.
     */
    public void testConstructingFromOneClass() throws Exception
    {
	inputString = 
	    "<project>" +
	    "  <class name=\"first class\"> " +
	    "  </class> " + 
	    "</project>" ;

	readerTestee = createGraphReaderTestee();

	result = readerTestee.readGraphFromXML( new ByteArrayInputStream(inputString.getBytes())) ;
	
	assertTrue(result.getVerticesCount()==1) ;
	assertNotNull( result.getNodeWithName("first class")) ;
	assertSame   ( result.getNodeWithName("first class"),
		       result.cloneVertices().iterator().next() ) ;
	
    } // EndMethod testConstructingFromOneClass

    /**
     * This method tests if the construction of a complex Directed Graph
     * from an XML output produced with Hydrogen works correctly.
     */
    public void testConstructingComplexGraphFromXML()
	throws Exception
    {
	inputString = 
	    "<project>" +
	    "  <class name=\"complex class\">"           +
	    "    <link kind           = \"superclass\" "    +
	    "          target         = \"first class\" />" +
	    "    <link kind           = \"usage\" " + 
	    "          name           = \"b\" "     + 
	    "          target         = \"B\" "     +
	    "          flatType       = \"B\" "      +
	    "          referenceDepth = \"0\" />"  +
	    "  </class> " +

	    "  <class name=\"first class\"> " +
	    "  </class> " +

	    "  <class name=\"B\"> " +
	    "    <link kind           = \"usage\" "         + 
	    "          name           = \"container\" "     + 
	    "          target         = \"complex class\" " +
	    "          template       = \"containerTemp&lt;complex class>\" " +
	    "          referenceDepth = \"1\" /> " +
	    "  </class> " + 
	    "</project>" ;

	readerTestee = createGraphReaderTestee();

	result = readerTestee.readGraphFromXML( new ByteArrayInputStream(inputString.getBytes())) ;
	
	assertTrue(result.getVerticesCount()==3) ;
	NamedVertex firstClass   = result.getNodeWithName("first class"  ) ;
	NamedVertex complexClass = result.getNodeWithName("complex class") ;
	NamedVertex bClass       = result.getNodeWithName("B") ;
	
	assertNotNull( firstClass   ) ;
	assertNotNull( complexClass ) ;
	assertNotNull( bClass       ) ;

	NamedEdge inheritLink = (NamedEdge)result.getEdge(complexClass,firstClass) ;
	NamedEdge usageLink1  = (NamedEdge)result.getEdge(complexClass,bClass) ;
	NamedEdge usageLink2  = (NamedEdge)result.getEdge(bClass,complexClass) ;

	assertNotNull( inheritLink ) ;
	assertNotNull( usageLink1  ) ;
	assertNotNull( usageLink2  ) ;

	java.util.Set expectedLinksFromComplexClass = new java.util.HashSet() ;
	expectedLinksFromComplexClass.add( inheritLink ) ;
	expectedLinksFromComplexClass.add( usageLink1  ) ;

	java.util.Set actualLinksFromComplexClass 
	    = new java.util.HashSet(result.getOutgoingEdges(complexClass) );
			  
	assertEquals( "LinksFromComplexClass",
		      expectedLinksFromComplexClass, 
		      actualLinksFromComplexClass
		      ) ;

	PropertyBag propOfInheritLink = inheritLink.getPropBag() ;

	assertEquals( "superclass" ,propOfInheritLink.getValueForProperty("kind")) ;

	PropertyBag propOfUsageLink2 = usageLink2.getPropBag() ;

	assertEquals( "usage",     propOfUsageLink2.getValueForProperty("kind")) ;
	assertEquals( "container", propOfUsageLink2.getValueForProperty("name")) ;
	assertEquals( "1",         propOfUsageLink2.getValueForProperty("referenceDepth")) ;
	assertEquals( "containerTemp<complex class>", 
		      propOfUsageLink2.getValueForProperty("template")
		      ) ;

    } // EndMethod testConstructingComplexGraphFromXML
    
    /**
     * Tests that adding a graph from an XML representation
     * to another graph correctly manages duplicate classes.
     */
    public void testAddGraphFromXMLToGraphInContext() throws Exception {
	String inputString1 = 
	    "<project>" +
	    "  <class name=\"A\" a=\"1\" > " +
	    "    <link target = \"B\" />" +
	    "  </class> " + 
	    "  <class name=\"B\"> " +
	    "  </class> " + 
	    "</project>" ;

	String inputString2 = 
	    "<project>" +
	    "  <class name=\"A\" b=\"4\" > " +
	    "    <link target = \"C\" />" +
	    "  </class> " + 
	    "  <class name=\"C\"> " +
	    "  </class> " + 
	    "</project>" ;

	readerTestee = createGraphReaderTestee();

	DirectedGraphWithDict result1  
	    = new DirectedGraphWithDictImpl() ;
	DirectedGraphWithDict context 
	    = new DirectedGraphWithDictImpl() ;

	readerTestee.readGraphFromXML
	    ( result1, context,  new ByteArrayInputStream(inputString1.getBytes()) ) ;
	
	assertEquals(result1,context) ;
	
	readerTestee = createGraphReaderTestee();
	DirectedGraphWithDict result2 
	    = new DirectedGraphWithDictImpl() ;
	readerTestee.readGraphFromXML
	    ( result2, context, new ByteArrayInputStream(inputString2.getBytes()) ) ;
	
	// Checking that through the use of the context graph, all
	// the NamedVertex with name 'A' in the three graph refer to the
	// same unique NamedVertex object.
	
	NamedVertex A1       = result1.getNodeWithName("A") ;
	NamedVertex A2       = result2.getNodeWithName("A") ;
	NamedVertex Ac       = context.getNodeWithName("A") ;

	assertTrue(A1!=null) ;
	assertSame(A1,A2) ;
	assertSame(A1,Ac) ;

	// Checking that the property of this NamedVertex object have been
	// correctly updated from the 2 source files.

	assertEquals("4",A1.getPropBag().getValueForProperty("b")) ;
	assertEquals("1",A1.getPropBag().getValueForProperty("a")) ;

    } // EndMethod testAddGraphFromXMLToGraphInContext

    /**
     * Tests that if 2 link representation $L1 and $L2 in an XML
     * graph representation have the same name, source and sink,
     * they are merged together in a unique link L, which
     * posses the attributes of L1, L2. In case an attribute
     * is present twice, if the link possesses the property 
     * "CONCATENATE"+nameOfThePropertyToConcatenate
     * the resulting attribute of L should be the concatenation of them.
     */
    public void testMergingAttributesDuplicateLinks() 
	throws Exception
    {
	// Graph without CONCATENATE option

	String inputString = 
	    "<project>" +
	    "  <class name=\"A\" > " +
	    "    <link target = \"B\" myAttri = \"TEST\" extra2 = \"Y\" />" +
	    "  </class> " + 
	    "  <class name=\"B\"> " +
	    "  </class> " + 
	    "  <class name=\"A\" > " +
	    "    <link target = \"B\" myAttri = \"TEST2\" extra1 = \"X\" />" +
	    "  </class> " + 
	    "</project>" ;
	
	readerTestee = createGraphReaderTestee();

	DirectedGraphWithDict result ;
	result = readerTestee.readGraphFromXML( new ByteArrayInputStream(inputString.getBytes())) ;

	NamedVertex a  = result.getNodeWithName("A") ;
	NamedVertex b  = result.getNodeWithName("B") ;
	
	assertNotNull( a ) ;
	assertNotNull( b ) ;

	NamedEdge theLink = (NamedEdge)result.getEdge(a,b) ;

	assertNotNull( theLink ) ;

	PropertyBag propOfLink = theLink.getPropBag() ;
	String myAttri =  propOfLink.getValueForProperty("myAttri") ;
	String extra2  =  propOfLink.getValueForProperty("extra2" ) ;
	String extra1  =  propOfLink.getValueForProperty("extra1" ) ;
	assertNotNull( myAttri ) ;
	assertNotNull( extra2  ) ;
	assertNotNull( extra1  ) ;
	assertTrue( "myAttri should either equals TEST or TEST2",
		myAttri.equals("TEST") || myAttri.equals("TEST2") ) ;
	assertEquals("Y",extra2) ;
	assertEquals("X",extra1) ;

	// Graph with CONCATENATE option
 
	inputString = 
	    "<project>" +
	    "  <class name=\"A\" > " +
	    "    <link target = \"B\" myAttri = \"TEST\" extra2 = \"Y\" CONCATENATEmyAttri=\"%\" />" +
	    "  </class> " + 
	    "  <class name=\"A\" > " +
	    "    <link target = \"B\" myAttri = \"TEST\" extra2 = \"Y\" CONCATENATEmyAttri=\"%\" />" +
	    "  </class> " + 
	    "  <class name=\"B\"> " +
	    "  </class> " + 
	    "  <class name=\"A\" > " +
	    "    <link target = \"B\" myAttri = \"TEST2\" extra1 = \"X\" CONCATENATEmyAttri=\"%\" />" +
	    "  </class> " + 
	    "  <class name=\"A\" > " +
	    "    <link target = \"B\" myAttri = \"TEST3\" />" +
	    "  </class> " + 
	    "  <class name=\"A\" > " +
	    "    <link target = \"B\" myAttri = \"TEST\" extra2 = \"Y\" CONCATENATEmyAttri=\"%\" />" +
	    "  </class> " + 
	    "</project>" ;

	readerTestee = createGraphReaderTestee();

	result = readerTestee.readGraphFromXML( new ByteArrayInputStream(inputString.getBytes())) ;

        a  = result.getNodeWithName("A") ;
        b  = result.getNodeWithName("B") ;
	
	assertNotNull( a ) ;
	assertNotNull( b ) ;

	theLink = (NamedEdge)result.getEdge(a,b) ;
	assertNotNull( theLink ) ;
	
	propOfLink = theLink.getPropBag() ;
	myAttri =  propOfLink.getValueForProperty("myAttri") ;
	extra2  =  propOfLink.getValueForProperty("extra2" ) ;
	extra1  =  propOfLink.getValueForProperty("extra1" ) ;
	String target  
	        =  propOfLink.getValueForProperty("target" ) ;
	assertEquals("Only the property myAttri should be concatenated.","B",target);
	assertNotNull( myAttri ) ;
	assertNotNull( extra2  ) ;
	assertNotNull( extra1  ) ;
	assertEquals( "myAttri should contains TEST%TEST2%TEST3",
		      "TEST%TEST2%TEST3",myAttri ) ;
	assertEquals("Y",extra2) ;
	assertEquals("X",extra1) ;

    } // EndMethod testMergingAttributesDuplicateLinks

    /**
     * Test that the reader records the oder of appearance of links
     * in an attribute call "oderOfFirstAppearance".
     */
    public void testOrderOfAppearance() throws Exception {
	String inputString = 
	    "<project>" +
	    "  <class name=\"A\" > " +
	    "    <link target = \"B\" name = \"Link1\" chainID=\"yes1\" />" +
	    "  </class> " + 
	    "  <class name=\"B\"> " +
	    "  </class> " + 
	    "  <class name=\"A\" > " +
	    "    <link target = \"B\" name = \"Link2\" chainID=\"yes2\" />" +
	    "    <link target = \"B\" name = \"Link1\" chainID=\"yes3\" />" +
	    "  </class> " + 
	    "</project>" ;
	
	readerTestee = createGraphReaderTestee();

	DirectedGraphWithDict result ;
	result = readerTestee.readGraphFromXML( new ByteArrayInputStream(inputString.getBytes()) ) ;

	NamedVertex a  = result.getNodeWithName("A") ;
	NamedVertex b  = result.getNodeWithName("B") ;
	
	NamedEdge link1 = result.getLinkWithSourceSinkName(a,b,"Link1") ;
	NamedEdge link2 = result.getLinkWithSourceSinkName(a,b,"Link2") ;

	String callOrder ;
	callOrder = link1.getPropBag().getValueForProperty("callOrder") ;
	assertEquals("callOrder of link1","1",callOrder);
	callOrder = link2.getPropBag().getValueForProperty("callOrder") ;
	assertEquals("callOrder of link2","2",callOrder);

	assertNull(link1.getPropBag().getValueForProperty("chainID")) ;
	assertNull(link2.getPropBag().getValueForProperty("chainID")) ;
    } // EndMethod testOrderOfAppearance

    /**
     * Tests that the reader records the oder of appearance in a recursive
     * manner when several incoming calls occurs.
     */
    public void testOrderOfAppearance2() throws Exception {
	String inputString = 
	    "<project>" +
	    "  <class name=\"C\" > " +
	    "    <link target = \"D\" name = \"CD\" chainID=\"1\" />" +
	    "  </class> " + 
	    "  <class name=\"A\" > " +
	    "    <link target = \"C\" name = \"AC\" chainID=\"1\" />" +
	    "  </class> " + 
	    "  <class name=\"C\" > " +
	    "    <link target = \"E\" name = \"CE\" chainID=\"2\" />" +
	    "  </class> " + 
	    "  <class name=\"B\" > " +
	    "    <link target = \"C\" name = \"BC\" chainID=\"2\" />" +
	    "  </class> " + 
	    "  <class name=\"C\" > " +
	    "    <link target = \"D\" name = \"CD\" chainID=\"3\" />" +
	    "  </class> " + 
	    "  <class name=\"B\" > " +
	    "    <link target = \"C\" name = \"BC\" chainID=\"3\" />" +
	    "  </class> " + 
	    "</project>" ;
	
	readerTestee = createGraphReaderTestee();

	DirectedGraphWithDict result ;
	result = readerTestee.readGraphFromXML( new ByteArrayInputStream(inputString.getBytes()) ) ;

	NamedVertex a  = result.getNodeWithName("A") ;
	NamedVertex b  = result.getNodeWithName("B") ;
	NamedVertex c  = result.getNodeWithName("C") ;
	NamedVertex d  = result.getNodeWithName("D") ;
	NamedVertex e  = result.getNodeWithName("E") ;

	NamedEdge ac = result.getLinkWithSourceSinkName(a,c,"AC") ;
	NamedEdge cd = result.getLinkWithSourceSinkName(c,d,"CD") ;
	NamedEdge bc = result.getLinkWithSourceSinkName(b,c,"BC") ;
	NamedEdge ce = result.getLinkWithSourceSinkName(c,e,"CE") ;

	String incomingOrder ;
	incomingOrder = ac.getPropBag().getValueForProperty("incomingOrder") ;
	assertEquals("incomingOrder of ac","a",incomingOrder) ;
	incomingOrder = bc.getPropBag().getValueForProperty("incomingOrder") ;
	assertEquals("incomingOrder of bc","b",incomingOrder) ;
	incomingOrder = cd.getPropBag().getValueForProperty("incomingOrder") ;
	assertEquals("incomingOrder of cd","a",incomingOrder) ;
	incomingOrder = ce.getPropBag().getValueForProperty("incomingOrder") ;
	assertEquals("incomingOrder of ce","a",incomingOrder) ;

	String callOrder ;
	callOrder = ac.getPropBag().getValueForProperty("callOrder") ;
	assertEquals("callOrder of ac","1",callOrder) ;
	callOrder = bc.getPropBag().getValueForProperty("callOrder") ;
	assertEquals("callOrder of bc","1",callOrder) ;
	callOrder = cd.getPropBag().getValueForProperty("callOrder") ;
	assertEquals("callOrder of cd","a.1,b.3",callOrder) ;
	callOrder = ce.getPropBag().getValueForProperty("callOrder") ;
	assertEquals("callOrder of ce","b.2",callOrder) ;

    } // EndMethod testOrderOfAppearance2

    
} // EndClass TestXMLGraphReader
