/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * Command.java
 * 
 * Time-stamp: <2003-03-11 11:27:10 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: Command.java,v 1.4 2004/09/09 19:35:46 ftaiani Exp $
 * 
 */
package laas.friends.cosmopen.interpreter  ;

import java.util.* ;

/**
 * This class encapsulates a command of a command
 * interpreter.
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.4 $
 * @since Version 0.0.2
 */
public abstract class Command extends AbstractDocumentedItem {

    private int    _maxNumbOfArgs ;
    private int    _minNumbOfArgs ;

    public int getMaxNbArgs() {
	return _maxNumbOfArgs ;
    } // EndMethod getMaxNbArgs

    /**
     * common part to all constructors. This might seem weird to define
     * an <code>init</code> method with only one line, but it is
     * pratical to have one single point of change for anything
     * common to all constructor of this class. (Planning
     * for changes...)
     */
    private void init(String aCommandName) {
	addAlias(getName()) ;
    } // EndMethod init

    public Command(String aCommandName) 
    {
	super(aCommandName) ;
	this.init(aCommandName) ;
	_maxNumbOfArgs = 0 ;
	_minNumbOfArgs = 0 ;
    } // EndConstructor Command

    /**
     * Constructs a new command and sets the min and
     * maximum number of arguments which this command
     * takes. Those numbers are used in 
     * {@link #thereIsRightNumberOfArgs(String[] args) } but
     * if this last method is actually called in the implementation of
     * {@link #handleCall(String[] args) } is the choice of
     * the command implementer.<br>
     * You can use the value <code>-1</code> 
     * for the max to indicate infinite.
     * It is a good idea to directly set the short and long
     * description of the command in its constructor using
     * {@link #setShortDescription(String aShortDescription) } and
     * {@link #setLongDescription(String aLongDescription) }.
     *
     * @see  #thereIsRightNumberOfArgs(String[] args)
     * @see  #handleCall(String[] args)
     * @see  #setShortDescription(String aShortDescription)
     * @see  #setLongDescription(String aLongDescription)
     */
    public Command(String aCommandName, 
		   int    minNumbOfArgs,
		   int    maxNumbOfArgs) 
    {
	super(aCommandName) ;
	this.init(aCommandName) ;
	_maxNumbOfArgs = maxNumbOfArgs ;
	_minNumbOfArgs = minNumbOfArgs ;
    } // EndConstructor Command

    /**
     * Tests in the number of arguments in <code>args</code>
     * is in the range given at construction. In case it
     * isn't an appropriate error message is printed on
     * the default console.<br>
     * @return <code>true</code> if the number of args is OK,
     *         <code>false</code> otherwise.
     * @see #handleCall(String[] args)
     */
    public boolean thereIsRightNumberOfArgs(String[] args) {
	// We must take into consideration that args[0] is
	// actually the name of the command, and that the
	// arguments of the command *itself* only go
	// from args[1] to args[args.length-1]

	boolean tooFewArgs  = (args.length < _minNumbOfArgs+1) ;
	boolean tooManyArgs = (args.length > _maxNumbOfArgs+1) && 
	    (_maxNumbOfArgs!=-1) ; // -1 == infinite

	boolean isOK = !tooManyArgs && !tooFewArgs ;

	if (tooManyArgs) {
	    println("Too many arguments for '"+this.getName()+"'") ;
	    println("Type 'help "+this.getName()+"' for help.") ;
	} ;

	if (tooFewArgs) {
	    println("Too few arguments for '"+this.getName()+"'") ;
	    println("Type 'help "+this.getName()+"' for help.") ;
	} ;

	return isOK ;
    } // End thereIsRightNumberOfArgs
    
    public String getLongDescription() {
	if (getAliasES().size()>1) {
	    StringBuffer aliasList = new StringBuffer("Aliases: ") ;
	    boolean thereIsAPrecedingAlias = false ;
	    for(Iterator i = getAliasES().iterator() ; i.hasNext() ; ) {
		if (thereIsAPrecedingAlias) aliasList.append(", ") ;
		aliasList.append((String)i.next()) ;
		thereIsAPrecedingAlias = true ;
	    } // EndFor Loop through aliases
	    return super.getLongDescription() + "\n" + aliasList ;
	}
	return super.getLongDescription() ;
    } // EndMethod getLongDescription

    /**
     * This methods contains the actual definition of the command. For
     * concrete commands it has to be redefined by the programmer.
     * It is a good pratice to test {@link #thereIsRightNumberOfArgs(String[]) }
     * just at the beginning of handleCall, and to abort method execution
     * in case the number of args is wrong.
     * @param args the arguments of the command. Note that args[0] is
     *             the name of the command, the actual arguments - if any -
     *             only begin at args[1].
     * @return     <code>true</code> if the command has been successfully executed
     *             <code>false</code> otherwise.
     * @see #thereIsRightNumberOfArgs(String[])
     */
    public abstract boolean handleCall(String[] args) throws Exception ;

    /**
     * The same as {@link #handleCall(String[])}, except that possible
     * internal exception are caught and a generic "INTERNAL ERROR" message
     * is printed. The program is not stopped.
     */
    public boolean handleCallCatchException(String[] args) 
	throws InterpreterException
    {
	try {
	    return handleCall(args) ;
	} 
	catch (InterpreterException interpreterException) {
	    throw (interpreterException) ;
	} // EndCatch InterpreterException

	catch (Exception except) {
	    println( "INTERNAL ERROR: " + except ) ;
	    except.printStackTrace() ;
	} // EndCath
	return false ;
    } // EndMethod handleCallCatchException

    private CommandListener _aListener = null ;

    /**
     * registers a {@link CommandListener } for the command. There can only
     * be one {@link CommandListener } per {@link Command } instance. This
     * listener will be notified each time the {@link Command } instance
     * will be successfully executed ({@link Command#handleCall(String[] args)
     * handleCall(args)} returns true) by a {@link CommandInterpreter }.
     */
    public void registerCommandListener(CommandListener listener) {
	this._aListener = listener ;
    } // EndMethod registerCommandListener

    /**
     * notifies the registered {@link CommandListener } if any
     * that the command has been successfully executed with
     * the given arguments. This method is called by the
     * {@link CommandInterpreter }.
     */
    public void notifyCommandListener(String[] args) {
	if (this._aListener!=null) 
	    this._aListener
		.aListenedCommandHasBeenSuccessfullyCalled(this,args) ;
    } // EndMethod getCommandListener

    SortedSet _aliasES = new TreeSet() ;

    public void addAlias(String anAlias) {
	_aliasES.add(anAlias) ;
    }

    /**
     * The name is treated as an alias as well, and will be returned.
     */
    public SortedSet getAliasES() {
	return _aliasES ;
    }

    /** ****************************************************
     * Uses {@link CommandInterpreter#out}. Here for convenience
     * purposes.
     */
    static public void println(String aLine) {
	CommandInterpreter.out.println(aLine) ;
    }

    public String toString() {
	return getName() ;
    }

} // EndClass Command
