/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * CommandInterpreter.java
 * 
 * Time-stamp: <2003-03-11 11:01:50 ftaiani>
 * @author Francois Taiani   <taiani@laas.fr>
 * CVS Info: $Id: CommandInterpreter.java,v 1.11 2008/10/30 14:32:19 ftaiani_normal Exp $
 * 
 */
package laas.friends.cosmopen.interpreter  ;

import laas.friends.cosmopen.utils.* ;

import java.io.* ;
import java.util.* ;

/**
 * This class acts as a repository of the possible
 * commands in a command line interpreter and is responsible
 * for reading commands from the standard input line
 * and for interpreting them.
 * @see Command
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.11 $
 * @since Version 0.0.2
 */
public class CommandInterpreter {

    /** ****************************************************
     * Instances of {@link CommandInterpreter} and of {@link Command}
     * use it to print things out. Can be overwritten for debug
     * purpose.
     */
    static public PrintStream out = System.out ;

    private String  _nameOfInterpreter ;

    // Commands are organized in groups. An Item may be
    // either a group or a command.

    private java.util.SortedMap _knownCommandS = new java.util.TreeMap();
    private java.util.SortedMap _knownItemS    = new java.util.TreeMap();
 
    private GroupOfDocumentedItemS _groupGeneral ;
    private GroupOfDocumentedItemS 
	_topGroupOfDocumentedItems = new GroupOfDocumentedItemS("Top Help Menu:") ;

    protected PropertyBag interpreterProperties = new PropertyBag() ;

    /** **************************************************** */
    public CommandInterpreter(String interpreterName) {

	_nameOfInterpreter = interpreterName ;
	_knownItemS.put(_topGroupOfDocumentedItems.getName(), _topGroupOfDocumentedItems) ; 

	_groupGeneral = new GroupOfDocumentedItemS("general") ;
	_groupGeneral.setShortDescription("generic commands") ;
	_groupGeneral.setLongDescription ("This group contains general commands that are "+
					  "related to the use of the interpreter itself." ) ;
	addGroupToKnownOnes(_groupGeneral) ;
	addCommandToKnownOnes( new DefaultHelpCommand(this)) ;
	addCommandToKnownOnes( new DefaultQuitCommand(this)) ;
	addCommandToKnownOnes( new DefaultRun_Command(this)) ;
	addCommandToKnownOnes( new DefaultShellExecuteCommand(this)) ;
	addCommandToKnownOnes( new DefaultSetOptionCommand   (this)) ;
	addCommandToKnownOnes( new DefaultUnsetOptionCommand (this)) ;

    } // EndConstructor CommandInterpreter
    
    /** ****************************************************
     * Commands are organized in groups for documentation purposes.
     * This returns the top group.
     * @see Command
     * @see AbstractDocumentedItem
     * @see GroupOfDocumentedItemS
     */
    public GroupOfDocumentedItemS getTopDocumentationGroup() {
	return _topGroupOfDocumentedItems ;
    } // EndMethod 


    /** ****************************************************
     * {@link Command}s may be grouped together into 
     * {@link GroupOfDocumentedItemS}, and these groups may
     * be recursively grouped to form a tree hierarchy.
     * This method adds a new group to the interpreter.
     */
    public void addGroupToKnownOnes( GroupOfDocumentedItemS aGroupOfItem ) {
	addItemToKnownOnesIntoGroup(aGroupOfItem,_topGroupOfDocumentedItems) ;
    } // EndMethod addGroupToKnownOnes


    /** ****************************************************
     * Similar to {@link #addCommandToKnownOnes(Command)} but
     * adds a "documented item" to a specific group.  The item can be 
     * a {@link Command} as well as a {@link GroupOfDocumentedItemS}.
     * If the group <code>aGroupOfItem</code> that is passed as second
     * argument is not known to the interpreter, an exception is raised.
     */
    public void addItemToKnownOnesIntoGroup( AbstractDocumentedItem anItem , 
					     GroupOfDocumentedItemS aGroupOfItem ) 
    {
	// I- The proposed group must be known.
	if (!_knownItemS.containsValue(aGroupOfItem))
	    throw new RuntimeException
		    ("Proposed group "+aGroupOfItem+" is unknown to interpreter." ) ;

	// II- CHECK FOR CONFLICTS
	// Note that the map _knownCommandS is included in
	// the map _knownItemS, so checking _knownItemS is sufficient

	for(Iterator i = anItem.getAliasES().iterator() ; i.hasNext() ; ) {
	    String alias = (String)i.next() ;
	    if (_knownItemS.containsKey(alias)) {
		// The alias has already been registered. It is OK
		// as long as the corresponding item is the same
		// as the one we want to register.
		if (_knownItemS.get(alias)!=anItem)
		    throw new RuntimeException
			("Conflicting aliases \""+alias+"\" between "+anItem+
			 " and "+_knownItemS.get(alias) ) ;
	    }
	} // EndFor Loop through aliases

	// III- NO CONFLICTS ADDING POSSIBLE

	boolean thisIsAcommand = anItem instanceof Command ;

	anItem.setContainingInterpreter(this);

	for(Iterator i = anItem.getAliasES().iterator() ; i.hasNext() ; ) {
	    String alias = (String)i.next() ;
	    _knownItemS   .put(alias, anItem) ; 
	    if (thisIsAcommand)
		_knownCommandS.put(alias, anItem) ; 
	} // EndFor Loop through aliases

	aGroupOfItem.addItemToGroup(anItem) ;

    } // End addItemToKnownOnesIntoGroup


    /** **************************************************** */
    public void addCommandToKnownOnes( Command aCommand ) {
	addItemToKnownOnesIntoGroup(aCommand, _groupGeneral ) ;
    } // EndMethod addCommandToKnownOnes


    /** **************************************************** */
    protected String[] breakStringIntoArgs(String rawInput) {
	String input = rawInput.trim() ;
	StringBuffer argUnderConstruction = new StringBuffer() ;
	java.util.Vector argSUnderConstruction = new java.util.Vector() ;
	if (input.length()==0) return new String[0] ;
	if (input.charAt(0)=='!') {
	    argSUnderConstruction.add("!") ;
	    if (input.length()>1) {
		String remainingString = input.substring(1).trim() ;
		if (input!="") argSUnderConstruction.add(remainingString) ;
	    }
	} // EndIf 1rst char is '!'
	else {
	    // LOOP ON ALL CHARACTER OF THE COMMAND LINE

	    input+=" "; // So that last argument does not need special care.
	    for(int i=0; i < input.length() ; i++) {

		// IF CHARACTER IS BLANK AND WE HAVE SOMETHING LOOKING LIKE AN ARG
		if ( (input.charAt(i)==' ') && 
		     (argUnderConstruction.length() > 0) ) {
		    argSUnderConstruction.add(argUnderConstruction.toString()) ;
		    argUnderConstruction = new StringBuffer() ;
		} // EndIf new arguments
		
		// IF NOT BLANK, THIS IS THE PART OF AN ARG
		if (input.charAt(i)!=' ') {
		    argUnderConstruction.append(input.charAt(i)) ;
		} // EndIf current char is not space
	    }
	    // END LOOP ON ALL CHARACTER OF THE COMMAND LINE
	}

	String[] dummy = new String[0] ;
	return (String[])argSUnderConstruction.toArray(dummy) ;

    } // EndMethod breakStringIntoArgs


    /** ****************************************************
     * interprets a stream of bytes as a flow of commands.
     */
    public void interpretCommandStream(java.io.InputStream input) 
	throws IOException, InterpreterException
    {
	BufferedReader reader = 
	    new BufferedReader( new  InputStreamReader (input) ) ;
	try {
	    while (reader.ready()) {
		interpretCommand(reader.readLine()) ;
	    } // EndWhile reader ready
	}
	catch (EOFException e) {}
    } // EndMethod interpretCommandStream


    /** **************************************************** */
    public void interpretCommand(String commandAndArgs)
	throws InterpreterException
    {
	String[] passedArgs = breakStringIntoArgs(commandAndArgs) ;
	if (passedArgs.length==0) return ; // Empty line does nothing
	SortedSet commandCandidatesS = completeCommandName(passedArgs[0]) ;
	if (commandCandidatesS.isEmpty()) // Given command is not known
	    throw new UnknownItemException
		("Command "+passedArgs[0]+" not known to CommandInterpreter.") ;
	if (commandCandidatesS.size()>1) // Too many possible completions
	    throw new AmbiguousInputException(commandCandidatesS) ;

	Command targetCommand = (Command)commandCandidatesS.first() ;
	// Replace possible alias with actual name. Useful if an history
	// facility is implemented.
	passedArgs[0] = targetCommand.getName() ;
	boolean success = targetCommand.handleCallCatchException(passedArgs) ;
	if (success) targetCommand.notifyCommandListener(passedArgs) ;
    } // EndMethod dispatchCommand

    /**
     * is used to notify the loop in {@link #launchInterpreter() }
     * that the call back method {@link #setStopRequestFlag(boolean) }
     * as been called ;
     */
    private boolean _requestForStopIssued = false ;

    /**
     * asks the user for command input on the standard
     * input line, and interactively execute them
     * according to the {@link Command } object
     * known to the interpreter.
     */
    public void launchInterpreter() {
        if (_requestForStopIssued) return ;
	CommandInterpreter.out.println(_nameOfInterpreter) ;
	CommandInterpreter.out.println
	    ("Type a command or help for a menu of available commands.") ;
	BufferedReader in
	    = new BufferedReader(new InputStreamReader(System.in));
	try {
	    while(!_requestForStopIssued) {
		CommandInterpreter.out.print("> ") ;
		try {
		    this.interpretCommand(in.readLine() ) ;
		} 
		catch (UnknownItemException exception) {
		    CommandInterpreter.out.println
			("Command no known. Type help for available commands.");
		} // EndCatch UnknownItemException
		catch (AmbiguousInputException ambiguous) {
		    CommandInterpreter.out.println
			("Ambiguous input. Possible completions are:");
		    Iterator iter = ambiguous.theCandidates.iterator() ;
		    while (iter.hasNext())
			CommandInterpreter.out.println(" - "+((AbstractDocumentedItem)iter.next()).getName()) ;
		} // EndCatch AmbiguousInputException
                catch (IllegalCommandArgException illegalArg) {
                    CommandInterpreter.out.println
			("Illegal argument. "+illegalArg.getMessage());
                }
	    } // EndWhile
        } catch (Exception except) {
            CommandInterpreter.out.println( "INTERNAL ERROR: " + except ) ;
            except.printStackTrace() ;
        } // EndCath
    } // EndMethod launchInterpreter


    /**
     * is called back by {@link DefaultHelpCommand } when
     * 'help' is entered with one argument. Print the
     * documentation of the command whose name is
     * passed, if known to the interpreter.
     */
    protected void printHelpForItem(String anItemName) 
	throws UnknownItemException, AmbiguousInputException {

	SortedSet itemCandidatesS = completeItemName(anItemName) ;
	if (itemCandidatesS.isEmpty()) // Given item is not known
	    throw new UnknownItemException
		("Item "+anItemName+" not known to "+this+".") ;
	if (itemCandidatesS.size()>1) // Too many possible completions
	    throw new AmbiguousInputException(itemCandidatesS) ;

	AbstractDocumentedItem targetItem = (AbstractDocumentedItem)itemCandidatesS.first() ;

	CommandInterpreter.out.println(targetItem.getHelpOutput()) ;

    } // EndMethod printHelpForItem

    /**
     * is called back by {@link DefaultHelpCommand} when
     * 'help' is entered with no argument. Print a menu
     * of all commands known to the interpreter.
     */
    protected void printHelpMenu() {
	CommandInterpreter.out.println(_topGroupOfDocumentedItems.getHelpOutput()) ;
    } // EndMethod printHelpMenu 

    /**
     * flip the flag {@link #_requestForStopIssued } to the passed value.
     * This is a callback method used by the {@link DefaultQuitCommand}
     * ('quit') to notify the launch loop of the interpreter that it
     * should stop.
     * @see #launchInterpreter()
     * @see #_requestForStopIssued
     */
    protected void setStopRequestFlag(boolean newValue) {
	_requestForStopIssued = newValue ;
    } // EndMethod setStopRequestFlag

    /**
     * returns the name of the interpreter.
     */
    public String toString() {
	return _nameOfInterpreter ;
    } // EndMethod toString

    /** ****************************************************
     * return the {@link AbstractDocumentedItem} (i.e. 
     * either a {@link Command} or a {@link GroupOfDocumentedItemS})
     * that carries the given name.
     */
    public AbstractDocumentedItem getItemForName(String anItemName) {
	return (AbstractDocumentedItem)_knownItemS.get(anItemName) ;
    } // EndMethod getItemForName

    public SortedSet completeItemName(String aPartialItemName ) {
	return completeNameFromSortedMap(aPartialItemName,_knownItemS) ;
    } // EndMethod completeItemName

    public SortedSet completeCommandName(String aPartialCommandName ) {

	return completeNameFromSortedMap(aPartialCommandName,_knownCommandS) ;

    } // EndMethod completeCommandName

    public SortedSet completeNameFromSortedMap( String aPartialItemName, 
						SortedMap dictionnary ) 
    {

	if (aPartialItemName==null)       return new TreeSet() ;
	if (aPartialItemName.length()==0) return new TreeSet() ;

	// We test if 'aPartialItemName' is directly associated with
	// an item. If yes, we return a SortedSet that only contains
	// this item.

	AbstractDocumentedItem itemAssociatedToPartialName =
	    (AbstractDocumentedItem)dictionnary.get(aPartialItemName) ;

	if (itemAssociatedToPartialName!=null) {
		SortedSet result = new TreeSet() ;
		result.add(itemAssociatedToPartialName) ;
		return result ;
	}  // EndIf itemAssociatedToPartialName!=null

	// 'aPartialItemName' is not directly associated with an
	// item. We extract a submap from 'dictionnary' that contains
	// all items that have keys that contains 'aPartialItemName'
	// as prefix.

	char[] nameAsArray = aPartialItemName.toCharArray() ;
	nameAsArray[nameAsArray.length-1]++ ;
	String lowerLimit = new String(nameAsArray) ;
	SortedMap range = dictionnary.subMap(aPartialItemName,lowerLimit ) ;
	
	return new TreeSet(range.values()) ;

    } // EndMethod completeNameFromSortedMap

    /**
     * Launches a shell command according to the passed string and
     * waits for its termination. The launch is delegated to
     * the java <code>Runtime</code> class. The only reason for this
     * method to be here is for test purposes: overloading it
     * allows observing which commands are (or are not) sent to
     * the shell in the JUnit test code.
     */
    protected void executeCommandAndPossiblyWait(String cmdString, boolean waitForTermination)
	throws IOException, InterruptedException 
    {
	
	Process launchedCommand =  Runtime.getRuntime().exec(cmdString) ;
	BufferedReader outputCommand
	    = new BufferedReader(new InputStreamReader(launchedCommand.getInputStream()));
	 
	String lastReadLine ;
	// Not sure if it works on all platforms / OS / JVM
	if (waitForTermination)
	    do {
		 lastReadLine = outputCommand.readLine() ;
		 if (lastReadLine!=null) CommandInterpreter.out.println(lastReadLine) ;
	    } while (lastReadLine!=null) ;

    } // EndMethdo executeCommand


    // **************************************************************
    // PRIVATE INNER CLASSES

    /** ****************************************************
     * Generic class for default commands. This class adds a
     * <code>_containingInterpreter</code> attribute to
     * the original {@link Command} class.
     */
    private abstract class DefaultCommand extends Command {
	protected CommandInterpreter _containingInterpreter ;

	public DefaultCommand(
			      String             aCommandName, 
			      CommandInterpreter aCommandInterpreter )
	{
	    super(aCommandName) ;
	    _containingInterpreter = aCommandInterpreter ;
	} // EndConstructor DefaultCommand

	public DefaultCommand(
			      String             aCommandName, 
			      CommandInterpreter aCommandInterpreter,
			      int    minNumbOfArgs,
			      int    maxNumbOfArgs)
	{
	    super(aCommandName,minNumbOfArgs,maxNumbOfArgs) ;
	    _containingInterpreter = aCommandInterpreter ;
	} // EndConstructor DefaultCommand

    } // EndPrivateClass DefaultCommand

    /** ****************************************************
     *  This quits the interpreter.
     */
    private class DefaultQuitCommand 
	extends DefaultCommand {
	public DefaultQuitCommand( CommandInterpreter aCommandInterpreter )
	{
	    super("quit",aCommandInterpreter) ;
	    this.setShortDescription("exits from "+aCommandInterpreter +"." ) ;
	    this.setLongDescription
		("'quit' stops "+ aCommandInterpreter +"." ) ;
	} // EndConstructor DefaultQuitCommand

	public boolean handleCall(String[] args) {
	    if (!thereIsRightNumberOfArgs(args)) return false;
	    _containingInterpreter.setStopRequestFlag(true) ;
	    return true ;
	} // EndMethod handleCall

    } // EndPrivateClass DefaultQuitCommand


    /** ****************************************************
     * Runs a serie of commands from a file.
     */
    private class DefaultRun_Command 
	extends DefaultCommand 
    {
	public DefaultRun_Command( CommandInterpreter aCommandInterpreter )
	{
	    super("run",aCommandInterpreter,1,1) ;
	    this.setShortDescription("Interpret a file as a flow of commands." ) ;
	    this.setLongDescription
		("'run <aFile>' reads a file and interprets each of its line as a command.") ;
	} // EndConstructor DefaultRun_Command

	public boolean handleCall(String[] args) throws InterpreterException {
	    if ( ! thereIsRightNumberOfArgs(args) ) return false ;

	    boolean executionOk = false ;

	    try {
		java.io.InputStream  input  = new java.io.FileInputStream (args[1]) ;
		
		CommandInterpreter.out.println
		    ("Starting interpretation of file "+args[1]+".");
		_containingInterpreter.interpretCommandStream(input) ;
		executionOk = true ;
		CommandInterpreter.out.println
		    ("Interpretation of file "+args[1]+" was successful.");

	    } catch (java.io.FileNotFoundException e) {
		CommandInterpreter.out.println( "ERROR WHILE READING: "+
				    "File '" + args[1] + "' could not be found.") ;
	    } catch (java.io.IOException e) {
		CommandInterpreter.out.println( "INTERNAL ERROR: " + e ) ;
		e.printStackTrace() ;
	    } // EndCatch

	    return executionOk ;
 	} // EndMethod handleCall

    } // EndPrivateClass DefaultRun_Command

    
    /** *****************************************************/
    private class DefaultHelpCommand 
	extends DefaultCommand {
	public DefaultHelpCommand( CommandInterpreter aCommandInterpreter )
	{
	    super("help",aCommandInterpreter,0,1) ;
	    addAlias("h");
	    addAlias("?");
	    this.setShortDescription("help menu or help on a command.") ;
	    this.setLongDescription
		("'help' with no arguments returns a "+
		 "menu of the available commands. \n"+
		 "'help <aCommandName>' returns the help on '<aCommandName>' " +
		 "if that command is known to "+ aCommandInterpreter +"." 
		 ) ;
	} // EndConstructor DefaultHelpCommand

	public boolean handleCall(String[] args) 
	    throws UnknownItemException, AmbiguousInputException 
	{
	    if (!thereIsRightNumberOfArgs(args)) return false ;

	    if (args.length == 1) {
		_containingInterpreter.printHelpMenu() ;
	    } // EndIf only one argument
	    
	    if (args.length == 2) {
		_containingInterpreter.printHelpForItem(args[1]) ;
	    } // EndIf only one argument

	    return true ;

	} // EndMethod handleCall

    } // EndPrivateClass DefaultHelpCommand

    /** ****************************************************
     *  This executes a shell command.
     */
    private class DefaultShellExecuteCommand 
	extends DefaultCommand {
	public DefaultShellExecuteCommand( CommandInterpreter aCommandInterpreter )
	{
	    super("!",aCommandInterpreter,1,1) ;
	    this.setShortDescription("run a shell command." ) ;
	    this.setLongDescription
		("'!' runs a shell command with its arguments. Which command are available "+
		 "depends on your underlying shell. The space between ! and the command can "+
		 "be ommited. ('!ls' on Unix print the content of the current directory." ) ;
	} // EndConstructor DefaultShellExecuteCommand

	public boolean handleCall(String[] args) throws Exception {
	    if (!thereIsRightNumberOfArgs(args)) return false ;
	    String   plainCommand ;
	    boolean  shouldExecuteWait ;
	    if (args[1].endsWith("&")) {
		plainCommand = args[1].substring(0,args[1].length()-1).trim() ;
		shouldExecuteWait = false ;
	    }
	    else {
		plainCommand = args[1] ;
		shouldExecuteWait = true ;
	    }
	    _containingInterpreter.executeCommandAndPossiblyWait
		(plainCommand,shouldExecuteWait) ;
	    return true ;
	} // EndMethod handleCall

    } // EndPrivateClass DefaultShellExecuteCommand

    private SortedMap knownOptionsAndTheirDocumentation = new TreeMap() ;

    public void addOptionToInterpreter(String optionName, String shortDocumentation) {
	knownOptionsAndTheirDocumentation.put(optionName, shortDocumentation) ;

    } // EndMethod addOptionToInterpreter

    public boolean optionIsSet(String optionName) {
	return interpreterProperties.hasProperty(optionName) ;
    } // EndMethod optionIsSet

    public void switchOptionOn(String optionName) {
	setOptionProperty(optionName,"");
    } // EndMethod switchOptionOn

    private void setOptionProperty(String optionName, String newValue) {
	this.interpreterProperties
	    .setPropertyToValue(optionName,newValue);
	OptionListener potentialListener 
	    = (OptionListener)optionNameToOptionListener.get(optionName) ;
	if (potentialListener!=null)
	    potentialListener.optionHasNewValue(optionName,newValue) ;
    } // EndMethod setOptionProperty

    private Map optionNameToOptionListener = new HashMap() ;

    /**
     * registers a listener for one of the options of the
     * interpreter. Beware : only one listener per option !
     *
     * @see #addOptionToInterpreter(String , String )
     * @see #optionIsSet(String)
     */
    public void setListenerForOption(OptionListener aListener,
				     String         optionName )
    {
	if (!knownOptionsAndTheirDocumentation.containsKey(optionName))
	    throw new RuntimeException("Unknown option.") ;
	optionNameToOptionListener.put(optionName,aListener) ;

    } // EndMethod setListenerForOption
    
    /** ****************************************************
     * abstract command to handle options
     */
    private abstract class AbstractSetUnsetCommand 
	extends DefaultCommand {

	public AbstractSetUnsetCommand( String             aCommandName, 
					CommandInterpreter aCommandInterpreter )
	{
	    super(aCommandName, aCommandInterpreter,1,1) ;
	} // EndConstructor DefaultCommand

	protected boolean abstractHandleCall(String[] args, String newValue) 
	    throws Exception {

	    if (!thereIsRightNumberOfArgs(args)) return false ;
	    if (!_containingInterpreter
		.knownOptionsAndTheirDocumentation
		.containsKey(args[1]) )
		throw new IllegalCommandArgException
		("'"+args[1]+"' is not a valid option.") ;
	    _containingInterpreter.setOptionProperty(args[1],newValue);
	    return true ;

	} // EndMethod abstractHandleCall

	public String getLongDescription() {

	    if (_containingInterpreter
		.knownOptionsAndTheirDocumentation
		.isEmpty()) 
		return super.getLongDescription() + "\nNo option is available." ;
	    else {
		StringBuffer result = new StringBuffer("\nHere are the availble options:\n");

		for ( java.util.Iterator i = 
			  _containingInterpreter
			  .knownOptionsAndTheirDocumentation
			  .keySet()
			  .iterator() ;
		      i.hasNext() ; ) 
		    {
			String optionName = (String)i.next() ;
			result.append
			    (" -"+" "+MiscUtils.ExtendWithSpaces(optionName,9)+" : ") ;
			if (_containingInterpreter.optionIsSet(optionName))
			    result.append("(on)  ") ;
			else
			    result.append("(off) ") ;
			result.append(
			     _containingInterpreter
			     .knownOptionsAndTheirDocumentation.get(optionName) + "\n"
			     ) ;
		    } // EndFor Loop through commands

		return super.getLongDescription() + result.toString();
	    } // EndElse
			
	} // EndMethod getLongDescription

    } // endprivateclass AbstractSetUnsetCommand

    /** ****************************************************
     * Switches an option on.
     */
    private class DefaultSetOptionCommand 
	extends AbstractSetUnsetCommand {
	public DefaultSetOptionCommand( CommandInterpreter aCommandInterpreter )
	{
	    super("set",aCommandInterpreter) ;
	    this.setShortDescription("turns on an option of the interpreter" ) ;
	    this.setLongDescription("'set' turns on an option of the interpreter." ) ;
	} // EndConstructor DefaultSetOptionCommand

	public boolean handleCall(String[] args) throws Exception {
	    return abstractHandleCall(args, "") ;
	} // EndMethod handleCall

    } // endprivateclass DefaultSetOptionCommand

    /** ****************************************************
     * Switches an option off.
     */
    private class DefaultUnsetOptionCommand 
	extends AbstractSetUnsetCommand {
	public DefaultUnsetOptionCommand( CommandInterpreter aCommandInterpreter )
	{
	    super("unset",aCommandInterpreter) ;
	    this.setShortDescription("turns off an option of the interpreter" ) ;
	    this.setLongDescription("'unset' turns off an option of the interpreter." ) ;
	} // EndConstructor DefaultUnsetOptionCommand

	public boolean handleCall(String[] args) throws Exception {
	    return abstractHandleCall(args, null) ;
	} // EndMethod handleCall

    } // endprivateclass DefaultUnsetOptionCommand

} // EndClass CommandInterpreter
