/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * PropertyBag.java
 * 
 * Time-stamp: <2003-03-10 16:39:45 ftaiani>
 *
 * CVS Info: $Id: PropertyBag.java,v 1.5 2004/10/27 17:16:53 ftaiani Exp $
 * @author Francois Taiani   <taiani@laas.fr>
 * 
 */
package laas.friends.cosmopen.utils  ;
import java.util.* ;

/**
 * This class is the base class of the {@link
 * laas.friends.cosmopen.graphAnalysis.NamedEdge} and {@link
 * laas.friends.cosmopen.graphAnalysis.NamedVertex} classes. Its main
 * feature is to support dynamic properties, which are stored as
 * "propertyName, propertyValue" pairs.<br> <b>Note:</b> This class is
 * redondant with {@link java.util.Properties} and could be replaced
 * by it in future released.
 *
 * @author Francois Taiani   <taiani@laas.fr>
 * @version $Name: release-1-3-rc1 $ $Revision: 1.5 $
 * @since Version 0.0.1
 */
public class PropertyBag {

    /**
     * contains the properties as (propertyName, propertyValue)
     * pairs.
     */
    protected final Hashtable _propertyTank ;

    public PropertyBag() {
	_propertyTank = new Hashtable() ;
    } // EndConstructor PropertyBag()

    public PropertyBag(PropertyBag toBeCopied) {
	_propertyTank = new Hashtable(toBeCopied._propertyTank ) ;
    } // EndConstructor PropertyBag()

    public Object clone() {
	return new PropertyBag(this) ;
    } // EndMethod clone

    public boolean hasProperty( String aPropertyName ) {
	String value = (String)_propertyTank.get(aPropertyName);
	return (value!=null) ;
    } // EndMethod hasProperty

    /**
     * returns true if the bag contains the passed property, and that property
     * is set to the passed value, false otherwise. In particular in the bag
     * does not contain that property, false is returned.
     */
    public boolean hasPropertyToValue( String aPropertyName, String aValue ) {
	String actualValue = (String)_propertyTank.get(aPropertyName);
	if (actualValue!=null)
	    return actualValue.equals(aValue) ;
	return false ;
    } // EndMethod hasPropertyToValue

    /**
     * As {@link #setPropertyToValue( String , String )} but do concatenation
     * when needed. The concatenation depends on the presence of 
     * a 'CONCATENATE' property in the property bag for
     * the considered property to be added ('CONCATENATE'+nameOfProperty).
     * If no, {@link #changePropertyToValue( String , String )} is as
     * {@link #setPropertyToValue( String , String )}.
     * If yes, the new value is concatenated to the old one. 
     * The 2 values are separated by the
     * value of the attribute CONCATENATE if it has one.
     */
    public void changePropertyToValue( String aPropertyName, 
				       String aValue ) 
    {
	if (!this.hasProperty(aPropertyName)) {
	     this.setPropertyToValue ( aPropertyName, aValue ) ;
	} // EndIf link1 has not this property.

	else {
	    boolean doConcatenate 
		= this.hasProperty("CONCATENATE"+aPropertyName) ;
	    if (doConcatenate) {
		String glueString
		    = this.getValueForProperty("CONCATENATE"+aPropertyName) ;
		String concatenatedProps 
		    = this.getValueForProperty(aPropertyName);
		String auxValue = aValue+glueString ;

		// The value to be added (aValue) may be a concatenation
		// of atomic values itself. We want to avoid doublon of
		// atomic values in the resulting string, so we slip
		// the added value in atomic string and add them one
		// after another if they're not already present.

		for(int index = 0 ; index < auxValue.length() ; ) {
		    int newIndex = auxValue.indexOf(glueString,index) ;
		    String newProp = auxValue.substring(index,newIndex) ;
		    index = newIndex + 1 ;
		    if (!isNewPropertyAlreadyPresent ( concatenatedProps, 
						       newProp, glueString ) ) 
			{
			    this.setPropertyToValue
				( aPropertyName, 
				  concatenatedProps + glueString + newProp ) ;
			} // EndIf !isNewPropertyAlreadyPresent
		} // EndFor index

	    } // EndIf doConcatenate
	} // EndIf link1 has this property
    } // EndMethod changePropertyToValue

    /**
     * When concatenating, checks if the property to be concatenated is already
     * present in the concatenation.
     * @see #changePropertyToValue( String , String )
     */
    private boolean isNewPropertyAlreadyPresent
	(String setOfProperties, 
	 String newProperty,
	 String concatenator) 
    {
	// To cope with the beginning and the end of the concatenation, we
	// use a trick: We search a bigger string, which has separators at
	// the beginning and at the end. The new property is searched with
	// concatenators before and after it as well.$

	String augmentedString = 
	    concatenator +
	    new String(setOfProperties) +
	    concatenator ;

	return (augmentedString.indexOf(concatenator+newProperty+concatenator)!=-1);

    } // EndMethod isNewPropertyAlreadyPresent

    /** ****************************************************
     * If aValue is null, the property is removed.
     */
    public void setPropertyToValue( String aPropertyName, 
				    String aValue ) 
    {
	if (aPropertyName!=null) {
	    if (aValue!=null)
		_propertyTank.put(aPropertyName,aValue) ;
	    else
		_propertyTank.remove(aPropertyName) ;
	}
    } // EndMethod setPropertyToValue
    
    public String getValueForProperty(String aPropertyName)
    {
	return (String)_propertyTank.get(aPropertyName) ;
    } // EndMethod getValueForProperty
    
    public void addBag(PropertyBag toBeAdded) {
	_propertyTank.putAll(toBeAdded._propertyTank) ;
    } // EndMethod addBag

    /**
     * returns an iterator on the properties keys contained
     * in the bag
     */
    public java.util.Iterator iterator() {
	return _propertyTank.keySet().iterator() ;
    } // EndMethod iterator

    public boolean equals(Object o) {
	if (!(o instanceof PropertyBag) ) return false ;
	PropertyBag autreBag = (PropertyBag)o;
	if (_relevantPropertiesForEquals==null)
	    return _propertyTank.equals(autreBag._propertyTank) ;
	// _relevantPropertiesForEquals is set, which means
	// only specific properties are to be taken into account during
	// the test.

	// LOOP ON RELEVANT PROPERTIES

	boolean equalSoFar = true ;
	for(int i=0 ; (i < _relevantPropertiesForEquals.length) && (equalSoFar) ; i++) {
	    String stringThis  = this    .getValueForProperty(_relevantPropertiesForEquals[i]) ;
	    String stringOther = autreBag.getValueForProperty(_relevantPropertiesForEquals[i]) ;
	    if (stringThis!=null)
		equalSoFar = stringThis.equals(stringOther) ;
	    else // stringThis is null
		equalSoFar = (stringOther==null) ;
	} // EndFor all relevant properties
	return equalSoFar ;
    } // EndMethod Equals

    public String toString() {
	if (_relevantPropertiesForEquals==null)
	    return "PropertyBag: "+ _propertyTank ; 
	StringBuffer result = new StringBuffer("PropertyBag: {") ;

	for(int i=0 ; (i < _relevantPropertiesForEquals.length) ; i++) {
	    String currentProp =  _relevantPropertiesForEquals[i] ;
	    if (hasProperty(currentProp))
		result
		    .append(currentProp)
		    .append("=")
		    .append(getValueForProperty(currentProp))
		    .append(", ");
	} // EndFor all relevant properties


// 	for(Iterator i = _propertyTank.keySet().iterator() ; i.hasNext() ; ) {
// 	    String currentProp = (String)i.next() ;
// 	    if (_relevantPropertiesForEquals.contains(currentProp) ) 
// 		if (hasProperty(currentProp))
// 		    result
// 			.append(currentProp)
// 			.append("=")
// 			.append(getValueForProperty(currentProp))
// 			.append(" ");
// 	} // EndFor



	result.append("}") ;
	return result.toString() ;
    } // EndMethod toString

    static private String[] _relevantPropertiesForEquals = null ;

    /**
     * changes the way property bags are tested for equality.  This
     * method should be used for debug and test purposes only. When
     * set, it specifies a set of properties that are used to test the
     * equality of property bags, i.e. only these properties will be
     * considered to decide if two bags are equals.<p>
     *
     * To go back to the default behavior (all properties are taken
     * into account) just call this method with a <code>null</code>
     * parameter.
     */
    static public void DebugSetPropertiesRelevantForEquality(String [] relevantProperties) {
	_relevantPropertiesForEquals = relevantProperties ;
    } // EndMethod debugSetPropertiesRelevantForEquality

    /**
     * returns the value of a property in a {@link
     * laas.friends.cosmopen.utils.PropertyBag} if it is non-null and
     * "" (the empty string) otherwise, instead of the null pointer.
     */
    public String provideProperty(String       popertyName) {

	String result = getValueForProperty(popertyName);
	if (result==null) return "" ;
	return result ;
    }

} // EndClass PropertyBag
