/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * monitor.c
 * 
 * Francois Taiani   <taiani@laas.fr>
 * Time-stamp: <2003-02-13 15:44:57 ftaiani>
 * 
 */
#include <errno.h>
#include <iostream.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>

#include "Monitor.h"

/**
 * Global variable. Very bad.
 */
int count_output = 0 ;

// -----------------------------------------------------

/**
 *  Used to recognize that a SIGPIPE was received.
 */
bool aSigpipeWasReceived = false ;

void handleSIGPIPE(int unknown1 , siginfo_t* someinfo, void* unknown2)
{
  aSigpipeWasReceived = true ;
  return ;
} // End handleSIGPIPE

// -----------------------------------------------------

Monitor::Monitor() 
{
  tracingIsActive = true ;
  struct sigaction monAction;

  monAction.sa_sigaction = handleSIGPIPE;
  sigemptyset(&monAction.sa_mask);
  monAction.sa_flags = 0;
  sigaction(SIGPIPE, &monAction, 0);

  childpid = 0 ; // No monitored process yet.

  clearLastReadLineS() ;
  prompt       = (char *)0 ; // null pointer
  flushingMark = (char *)0 ; // null pointer
} // End Constructor

// -----------------------------------------------------

char* Monitor::getPrompt() 
{
  return prompt ;
}

// -----------------------------------------------------

void Monitor::setPrompt(char * newPrompt) 
{
  prompt = newPrompt ;
}

// -----------------------------------------------------

char* Monitor::getFlushingMark() 
{
  return flushingMark ;
}

// -----------------------------------------------------

void Monitor::setFlushingMark(char * newFlushingMark) 
{
  flushingMark = newFlushingMark ;
}

// -----------------------------------------------------

void Monitor::perrorAndExit(char * msg) 
{
  perror(msg) ;
  exit(1) ;
} // End perrorAndExit

// -----------------------------------------------------

/**
 *  Is used to redirect file descriptors using close and dup.
 *  All file descriptors with IDs smaller than redirectedFd
 *  must be used for this to work !
 *  The fd redirectedFd is first closed, and then
 *  the targetFd file is duplicated.
 *  In case anything goes wrong, the program is terminated. 
 *  (Quite brutal !)
 */
void Monitor::redirectFileDToFileD(int redirectedFd, int targetFd) 
{
  int resultOfDup ;

  if (close ( redirectedFd )!=0) goto failed ; // close the fd to be redirected
  cerr << "close succeeded on " << redirectedFd << endl ;
  resultOfDup = dup ( targetFd ) ;             // connecting fd to fdToBeDuplicated

  if (resultOfDup!=redirectedFd) goto failed ;
  cerr << "dup succeeded on " << targetFd << endl ;

  return ;

  failed :
    {
      cerr << "Redirection failed." << endl ;
      exit(1) ;
    } // End If failure

} // End function redirectFileDToFileD

// -----------------------------------------------------

int Monitor::waitForEndOfMonitoredProcess()
{
  if (childpid<1) {
    cerr << "In Monitor::waitForEndOfMonitoredProcess()" << endl ;
    cerr << "No Monitored process to wait for !!" << endl ;
    return -1;
  }

  pid_t returnedPid ;
  do {
    errno = 0 ;
    returnedPid = waitpid(childpid,(int*)0,0) ;
    if (errno==EINTR) errno = 0 ; // Ignoring interrupt
  }
  while((returnedPid!=childpid) && (errno==0));
  
  return 0;

} // End waitForProcess

// -----------------------------------------------------

int Monitor::launchMonitoredProcess ( char *  pathToProgram )
{
  char * argv[1] ;
  argv[0] = pathToProgram ;
  launchMonitoredProcess(1, argv) ;
}

// -----------------------------------------------------

int Monitor::launchMonitoredProcess ( int           argc   , 
				      char *  argv[] )
{
  if (tracingIsActive)
    cerr << "launchMonitoredProcess: " << argv[0] << endl ;

  if (childpid>0) {
    cerr << "In Monitor::launchMonitoredProcess()" << endl ;
    cerr << "ERROR : A monitored process is already active!" << endl ;
    return -1;
  }

  int pipeToChild[2], pipeFromChild[2] ;

  if ( pipe(pipeToChild) < 0 || pipe(pipeFromChild) < 0 )
    perrorAndExit("Pipes could not be created.") ;

  childpid = fork() ;

  switch (childpid)
    {
    case -1 :
      perrorAndExit("Fork returned -1") ;

    case 0 :
      {
	// CHILD PROCESS
	cerr << "Child Process " << getpid() << endl ;

	close ( pipeToChild   [1]  ) ; // Close writing end
	close ( pipeFromChild [0]  ) ; // Close reading end

	// Redirect standard input (fd 0) to pipeToChild[0]
	redirectFileDToFileD(0,pipeToChild[0]) ;

	// Redirect standard output (fd 1) to pipeFromChild[1]
	redirectFileDToFileD(1,pipeFromChild[1]) ;

	// Redirect standard error output (fd 2) to pipeFromChild[1]
	redirectFileDToFileD(2,pipeFromChild[1]) ;

	// Because exec[something] does not use any arg count
	// parameter but instead uses a null argument to mark
	// the end of the arg list, we have to make a copy
	// of the original list, and add a terminating null arg.

	char ** argv_copy = (char**)calloc(argc+1, sizeof( char * ) ) ;
      
	memcpy(argv_copy, argv, argc*sizeof( char * )) ;
	argv_copy[argc] = 0 ;

	cerr << "Launching " << argv_copy[0] << endl ;

	if (execvp( argv_copy[0], argv_copy )!=0) 
	  {
	    cerr << "Call to " << argv_copy[0] << " failed." << endl ;
	    perrorAndExit("Aborting") ;
	  }
	// PROGRAM TERMINATION
	cerr << "exec( " << argv_copy[0] << " ) terminated anormally." << endl ;
	free(argv_copy);
	exit(1) ;
	break ;
      }
    default :
      // PARENT PROCESS
      // close ( pipeFromChild [1]  ) ; // Close writing end
      close ( pipeToChild   [0]  ) ; // Close reading end

      childin  = fdopen( pipeToChild[1], "w" ) ;
      if (childin  == NULL) perrorAndExit("Could no create stream childin.") ;
      setlinebuf( childin  ) ;
      writeToChildOut  = fdopen( pipeFromChild [1], "w" ) ;
      if (writeToChildOut  == NULL) 
	perrorAndExit("Could no create stream writeToChildOut.") ;
      setlinebuf( writeToChildOut  ) ;
      childoutfd = pipeFromChild [0] ;
      usleep(10) ;
      cerr << "Parent Process " << getpid() << endl ;
      break ;

    } // End switch

  return childpid ;

} // End launchMonitoredProcess

// -----------------------------------------------------

void Monitor::readChildOutUntilNextPrompt()
{
  if (prompt==0) {
    cerr << "In Monitor::readChildOutUntilNextPrompt() : " << endl ;
    cerr << "ERROR : Was called though not prompt was defined. " << endl 
	 << "Aborting" << endl ;
    exit(1) ;

  } // EndIf no prompt defined

  clearLastReadLineS() ;
  int nbTotalRead = 0 ;
  int lengthOfPrompt = strlen(prompt); // (O)ptimization: Not necessary to
                                       // compite it each time again.
  int lengthOfFMark = 0 ;
  if (flushingMark!=0)
    lengthOfFMark  = strlen(flushingMark) ;

  bool endsWithPrompt = false ;

  do {
    // Here we test that there is still enough place in the buffer
    // to store the next read character. We need one byte for the
    // next character to be read and one byte for the string
    // terminating byte "\0". That's the reason for + 2 (=1+1)
    
    if (nbTotalRead+2>sizeof(lastReadLineS)) {
      cerr << "In Monitor::readChildOutUntilNextPrompt" << endl 
	   << "ERROR: Buffer starvation. Aborting."     << endl ;
      exit(1) ;
    }
    nbTotalRead += 
      ReadInterruptSafe	( childoutfd, 
			  lastReadLineS+nbTotalRead, 
			  1
			  ) ;
    if ((lengthOfFMark>0) && (nbTotalRead>=lengthOfFMark))
      if (strcmp( lastReadLineS+nbTotalRead-lengthOfFMark, flushingMark )==0)
	{
	  cerr << "\""<<flushingMark << "\" encountered. Clearing input buffer." << endl ;
	  clearLastReadLineS() ;
	  nbTotalRead = 0 ;
	} // End If there is a flushing mark and we've just found it

    if (nbTotalRead>=lengthOfPrompt)
      endsWithPrompt = 
	( strcmp( lastReadLineS+nbTotalRead-lengthOfPrompt, prompt )==0 ) ;

  } while (!endsWithPrompt);
  
} // EndMethod readChildOutUntilNextPrompt

// -----------------------------------------------------

ssize_t Monitor::ReadInterruptSafe(int fd, char *buf, size_t count) {

  int  totalRead = 0 ;
  int  nbRead ;
  bool readWasInterrupted ;

  // READ LOOP AS LONG AS INTERRUPTED

  do {

    readWasInterrupted = false ;

    nbRead = read(fd, buf+totalRead ,count-totalRead) ;
    if (nbRead>0) {
      totalRead += nbRead ;
      buf[totalRead]=(char)0 ;
#undef cout
      cout << buf ;
#define cout coutdyn
    }
    else if ( (nbRead<0) && (errno==EINTR) ) {
      readWasInterrupted = true ;
    } else 
      perror("Error While Reading\n") ;

  } while (readWasInterrupted) ;

  // END OF READ LOOP

  return totalRead ;

} // End ReadInterruptSafe

// -----------------------------------------------------

void Monitor::readChildOutUntilEOF()
{
  fd_set  fdvar ;
  struct timeval timeout ;
  int nbReady = 0 ;
  int nbRead  = 0 ;
  int readWasInterrupted ;

  FD_ZERO( &fdvar) ;
  FD_SET ( childoutfd, &fdvar ) ;

  timeout.tv_sec  = 0 ; 
  timeout.tv_usec = 10000 ;

  clearLastReadLineS();

/*   printf("\n\n***********************\n"); */
/*   printf(    "In readChildOutUntilEOF\n"); */
/*   printf(    "***********************\n\n"); */

  /* DO WHILE LOOP STILL SOMETHING TO READ
   * We want to read from childoutfd and print the result
   * on the stdout as long as there is something to read
   * but we don't want to block in a read if no data is available
   * (which happens with pipes for instance) we use a select.
   * We also must account for the fact that the read may be
   * interrupted.
   */

  int nbTotalRead = 0 ;

  do {

    readWasInterrupted = 0 ;
    nbReady = select(childoutfd+1,&fdvar,(fd_set*)0,(fd_set*)0, &timeout);

    if (nbReady>0) 
      nbTotalRead += 
	ReadInterruptSafe ( childoutfd, lastReadLineS + nbTotalRead, 1 ) ;
    
  } while (nbReady>0) ;

  /* END OF DO WHILE STILL SOMETHING TO READ */

} // End readChildOutUntilEOF

// -----------------------------------------------------

/**
 * Sends a string with a terminating '\n' to a stream
 * and prints possible error. (Though
 * apparently writing to a stream never returns errors.)
 * That function also tests for SIGPIPE signals a posteriori
 * and abort a program if one was raised.
 */
void Monitor::sendLineToStream(FILE * stream, char * msg)
{
  aSigpipeWasReceived = false ;
  if (fprintf( stream,"%s\n", msg)<0)
      perrorAndExit("sendLineToStream failed.") ;

  if (aSigpipeWasReceived) 
    {
      cerr << "sendLineToStream(\"" << msg << "\") failed." << endl ;
      cerr << "A SIGPIPE was received. Aborting." << endl ;
      exit(1) ;
    }

} // End sendLineToStream

// -----------------------------------------------------

/**
 * Sends a string with a terminating '\n' to the standard input of the 
 * monitored process and prints possible error. (Though
 * apparently writing to a stream never returns errors. Adequate
 * signals are delivered instead like SIGPIPE in our case !)
 * message.
 */
void Monitor::sendLineToChild(char * msg)
{
  if (tracingIsActive)
    cerr << "sendLineToChild: " << msg << endl ;
  sendLineToStream(writeToChildOut, msg);
  sendLineToStream(        childin, msg);
  readChildOutUntilEOF() ;

} // End sendToChild

// -----------------------------------------------------

void Monitor::sendLineToChildBlocking(char * msg)
{
  if (tracingIsActive)
    cerr << "sendLineToChildBlocking: " << msg << endl ;
  sendLineToStream(writeToChildOut, msg);
  sendLineToStream(        childin, msg);
  readChildOutUntilNextPrompt() ;

} // End sendToChild

// -----------------------------------------------------

bool Monitor::lastReadLineSDoesContainRegexp(const char * regexp) {
  return matchRegexpInStringGetSubmatch(regexp,lastReadLineS,0,0) ;
} // EndMethod lastReadLineSDoesContainRegexp

// -----------------------------------------------------

bool Monitor::stringDoesContainRegexp( const char * stringToSearch,
				       const char * regexp ) {
  return matchRegexpInStringGetSubmatch(regexp,stringToSearch,0,0) ;
} // EndMethod doesStringContainRegexp
				       
// -----------------------------------------------------

bool Monitor::matchRegexpInStringGetSubmatch
         ( 
	  const char * regexp,
	  const char * stringToSearch,
	  size_t       maxNbSubstrings ,
	  regmatch_t * matchingSubstrings 
	  )
{
  if (tracingIsActive)
    cerr << "matchRegexpInStringGetSubmatch: " << regexp << "-> " << endl ;
  int errorcode ;
  regex_t preg ;
  char errmsg[1000] ;
  bool result = true ;

  // errorcode = regcomp(&preg, regexp, 0);
  errorcode = regcomp(&preg, regexp, REG_NEWLINE );
  if (errorcode!=0) 
    {
      regerror(errorcode,&preg,errmsg,1000) ;
      cerr << "regcomp failed : " << errmsg << endl ;
      return false ;
    }
  errorcode = regexec( &preg                ,
		       stringToSearch , 
		       maxNbSubstrings      , 
		       matchingSubstrings   , 
		       0 ) ;

  if (errorcode==REG_NOMATCH)
      result = false ;
  else if (errorcode!=0) 
    {
      regerror(errorcode,&preg,errmsg,1000) ;
      cerr << "regexec failed : " << errmsg << endl ;
      regfree(&preg);
      return false ;
    }
  regfree(&preg);
  if (tracingIsActive)
    cerr << "Result " << (result?"YES":"NO") << endl ;
  return result ;

} // End matchRegexpInStringFromOffsetGetSubmatch
  
// -----------------------------------------------------

char* Monitor::createSubstringFromLastReadLineS(int start, int end) {

  int length = end-start ;
  char * result = (char*)malloc(length+1) ;
  strncpy(result,lastReadLineS+start,length);
  result[length]='\0';
  return result ;

} // EndMethod createSubstringFromLastReadLineS

bool Monitor::interceptsTracingMode() {
  bool result = tracingIsActive ;
  tracingIsActive = false ;
  return result ;
} // EndMethod interceptsTracingMode


void Monitor::restoreTracingMode(bool interceptedTracing) {
  tracingIsActive = interceptedTracing ;
} // EndMethod interceptsTracingMode
