/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * monitor.h
 * 
 * Francois Taiani   <taiani@laas.fr>
 * Time-stamp: <2003-04-07 15:59:11 ftaiani>
 * 
 */
#ifndef MONITOR_H
#define MONITOR_H

#include <errno.h>
#include <regex.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>

/**
 * (D)angerous and Dirty: We overload cerr and cout in order
 * to automatically print a specific prompt along with
 * the output of the monitoring programm. In order for
 * that to work propertly monitor.h must come as last
 * include.
 */

extern int count_output ;

#define cerrdyn ( cerr << "(dynobs) " << ++count_output << ": " )
#define coutdyn ( cout << "(dynobs) " << ++count_output << ": " )
#define cerr cerrdyn
#define cout coutdyn

/**
 * Framework class to monitor a console base program from
 * another program.
 * (A) Attention: That class set up a signal handler and SIGPIPE
 * and does not clearly clean it on destruction. (Quik and dirty :(
 */

class Monitor {

public:

  /**
   * Constructor. Does some init work.
   */
  Monitor() ;

  /**
   * as waitpid but interupt safe.
   */
  int waitForEndOfMonitoredProcess() ;

  /**
   * launches the executable in pathToProgram
   * and redirect its standard input and output
   * in order for the monitoring functions (like sendLineToStream, 
   * and readChildOutUntilEOF) to work.
   * The Process ID of the new process is returned.
   */
  int launchMonitoredProcess(char * pathToProgram) ;

  /**
   * same as before but with an arg vector. argv[0] should
   * contain the command to be launched.
   */
  int launchMonitoredProcess ( int argc, char *  argv[] );

  /**
   * reads the standard output of the monitored process until
   * nothing comes out of it. (Like a flush.) In the present
   * version, what is read is printed of the standard output
   * of the monitoring process.
   */
  void readChildOutUntilEOF() ;

  void readChildOutUntilNextPrompt() ;

  /**
   * Sends a string with a terminating '\n' to the standard input of the 
   * monitored process and reads the following output of the monitored
   * process.  The reading is asynchronous and uses a time out
   * to decide that data is no more available. This is an important
   * limitation, because in case of slow communication, some
   * data may be lost. Use the sendLineToChildBlocking whenever
   * you can instead.
   *
   * The methods tries to detect possible errors due to the 
   * pipe communication . (Though apparently writing to a stream 
   * never returns errors. Adequate signals are delivered instead 
   * like SIGPIPE in our case !). If a SIGPIPE is detected, the program 
   * is aborted. The msg is also printed on the standard output of the
   * monitored process.
   */
  void sendLineToChild(char * msg) ;

  /**
   * As sendLineToChildBlocking, but the output of the child is read
   * until a prompt (as defined by setPrompt() ) is read, indicating
   * that the child is waiting for new input.
   */
  void sendLineToChildBlocking(char * msg) ;

  bool matchRegexpInStringGetSubmatch( 
	       const char * regexp,
	       const char * stringToSearch,
	       size_t       maxNbSubstrings ,
	       regmatch_t * matchingSubstrings ) ;

  bool stringDoesContainRegexp( const char * stringToSearch,
				const char * regexp ) ;

  bool lastReadLineSDoesContainRegexp(const char * regexp) ;

  /**
   * The prompt is a string, which is used by the monitoring process
   * to recognize when the monitored process awaits new inputs.
   * It is null by default.
   */
  char* getPrompt() ;
  void  setPrompt(char * newPrompt) ;

  /**
   * The FlushingMark is a string, which is used by the monitoring process
   * to recognize when to flush the lastReadLineS buffer between
   * 2 prompts. It is null by default and has no effect when null.
   * It should be set to avoid buffer starvation between 2
   * direct interaction with the monitored process, if the monitored
   * process produces a long output that may not fit in the
   * internal (protected) 'lastReadLineS' variable. (Which
   * in turn is used to analyse a posteriori the monitoree's output.)
   */
  char* getFlushingMark() ;
  void  setFlushingMark(char * newFlushingMark) ;

  /**
   * Extract a substring from the last read lines and put
   * it in a new created char * (with malloc). (A)ttention ! It is the
   * responsability of the caller to release the allocated
   * memory with a corresponding free(..) call on the
   * returned pointer.
   */
  char* createSubstringFromLastReadLineS(int start, int end) ;

  /**
   * [CHANGE] (Francois Taiani) Mon Nov 25 2002 15:28:51
   * When true, some of the stimuli sent to the monitored process
   * are echoed on the console. If false by default.
   */
  bool tracingIsActive ;

protected:

  /**
   * does the same as read, except that it is immune to interrupts,
   * and that errors are printed on stderr and not reported through
   * the return value.
   */
  static ssize_t ReadInterruptSafe(int fd, char *buf, size_t count) ;

  /** prompt of the monitored process. Is used to synchronize input and output */
  char * prompt ;

  /** flushingMark of the monitoree. It is used to synchronize input and output */
  char * flushingMark ;

  /** Pid of the monitor process */
  pid_t childpid ;

  /**
   *  Those variables are initialised by
   *  launchMonitoredProcess. They refer to the standard input
   *  and standard output of the child process.
   */

  FILE *childin ;
  FILE *writeToChildOut ; 

  /**
   * Used by the parent to write on the stdout of the child
   * Is useful when sending command to the monitored process
   * for this command to be printed on the stadout of that process
   */

  int   childoutfd ;

  void perrorAndExit(char * msg) ;
  /**
   *  Is used to redirect file descriptors using close and dup.
   *  All file descriptors with IDs smaller than fdToBeDuplicated
   *  must be used for this to work !
   *  The fd fdToBeDuplicated is first closed, and then
   *  the targetFd file is duplicated.
   *  In case anything goes wrong, the program is terminated. 
   *  (Quite brutal !)
   */
  void redirectFileDToFileD(int redirectedFd, int targetFd) ;

  /**
   * Contains the last bunch of line that were
   * read with readChildOutUntilEOF.
   */

#define LENGTHOFBUFFER 10000

  char lastReadLineS[LENGTHOFBUFFER] ;

  inline void clearLastReadLineS() {
    lastReadLineS[0]=(char)0 ;
  } // End ClearLastReadLineS

  /**
   * Sends a string with a terminating '\n' to a stream
   * and prints possible error. (Though
   * apparently writing to a stream never returns errors.)
   * That function also tests for SIGPIPE signals a posteriori
   * and abort a program if one was raised.
   */
  void sendLineToStream(FILE * stream, char * msg) ;

protected:
  /**
   * Put the tracing mode to 'false' (no tracing), and returns
   * the previous value of the tracing mode.
   * This method should be used when a Monitor class
   * makes nested invocation on itself to avoid
   * too many tracing messages.
   * The standard use would be:
   * <pre>
   * void MyClass::MyMethod() {
   *   bool interceptedTracing = interceptsTracingMode() ;
   *   ...
   *   ... // Calls to other Monitoring methods, which
   *   ... // normally produce tracing
   *   ...
   *   if (interceptedTracing) <traces> ;
   *   ...
   *   restoreTracingMode(interceptedTracing) ;
   * }
   * </pre>
   */
  bool interceptsTracingMode() ;
  /**
   * Restores an intercepted tracing mode.
   */
  void restoreTracingMode(bool interceptedTracing) ;

} ; // End Class Monitor

#endif // MONITOR_H
