/*
 * Copyright 2003 LAAS-CNRS
 *
 *   LAAS-CNRS
 *   7, avenue du Colonel Roche
 *   F-31077 Toulouse Cedex 4 / France
 *
 * Author: Francois Taiani
 *
 * This file is part of OPSBrowser.
 * 
 * OPSBrowser is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * OPSBrowser is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with OPSBrowser; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 */
/**
 * MonitorGDB.cpp
 * 
 * Francois Taiani   <taiani@laas.fr>
 * Time-stamp: <2003-03-05 17:53:04 ftaiani>
 * 
 */
#include <stdlib.h>
#include <string.h>
#include <iostream.h>

#include <sys/types.h>
#include <regex.h>

#include "MonitorGDB.h"


// -----------------------------------------------------------

/**
 * [CHANGE] (Francois Taiani) Mon Nov 25 2002 14:52:35
 * the extra arguments after argv[1] (note: argv[0] is the
 * executable name), are not sent any more to gdb, but to
 * the target program instead.
 */
MonitorGDB::MonitorGDB(int argc, char *  argv[] ) : Monitor() {
  argv[0] = "gdb" ;
  setPrompt("(gdb) ") ;

  char** extra_argv ;
  int    extra_argc ;

  // We distinguish the 2 cases: When more that 2 args are present, we
  // set the extra_argv pointer on the third arg and initialize
  // extra_argc accordingly otherwise we set extra_argv on a NULL
  // char* and extra_argc to null.

  if (argc > 2) {
    extra_argv = argv+2;
    extra_argc = argc-2;
    argc=2;
  }
  else {
    extra_argv = &NULL_STRING; 
    extra_argc = 0 ;
    argc = (1<argc?argc:1) ;
  }

  cerr << "Launching " ; 

  // LOOP ON ARGUMENTS

  for (int i=0 ; i < argc ; i++) {
    fprintf(stderr,"%s ", argv[i] ) ;
  } // EndFor
  fprintf(stderr,"\n") ;

  launchMonitoredProcess(argc, argv) ;
  readChildOutUntilNextPrompt() ;

  // There are extra args, that we tackle now.  We want we extra args
  // to be used as command line args for the programmed being
  // "debugged" by gdb.  For that we use the gdb command "set args"
  // and the inherited method "sendLineToChildBlocking".  We first
  // construct a string beginning with "set args" and containing all
  // the extra args.

  if (extra_argc>0) {
    // Big dirty hack... command_part[0] is the former argv[1], and
    // the first extra arg begins at command_part[1], i.e. the former
    // argv[2].
    char** command_part    = extra_argv-1 ;
    int    command_part_nb = extra_argc+1 ;
    command_part[0] = "set args" ; 
    int command_total_length = 0 ;

    // Computing the whole command length to do a clean memory
    // allocation
    for (int i=0; i < command_part_nb; i++) {
      command_total_length += strlen(command_part[i])+1 ;
    } // EndFor

    // THE memory allocation
    char * command_string = (char *)malloc(command_total_length+1);

    // Constructing the complete command string
    for (int i=0, cursor=0; i < command_part_nb; i++) {
      strcpy(command_string+cursor, command_part[i]);
      cursor += strlen(command_part[i]) ;
      command_string[cursor] = ' ' ;
      cursor ++ ;
    } // EndFor    

    // Delivering the string and destroying it afterward
    this->sendLineToChildBlocking( command_string ) ;
    free(command_string) ;

  } // EndIf there are some extra args
  else
    cerr << "No extra args found." << endl ;

} // End Constructor

void MonitorGDB::gdbSource( char * fileName ) {
  
  bool interceptedTracingMode = interceptsTracingMode() ;

  if (interceptedTracingMode) 
    cerr << "gdbSource: " << fileName << endl ;

  char * source = "source " ;
  char * commandString = (char *)malloc(strlen(fileName)+strlen(source)+1) ;
  strcpy(commandString,source) ;
  strcpy(commandString+strlen(source),fileName) ;
  sendLineToChildBlocking( commandString ) ;
  if (lastReadLineSDoesContainRegexp("Error in sourced command file") ) {
    cerr << "In MonitorGDB::gdbSource command is " << commandString << endl ;
    cerr << "Gdb failed to execute the file. Aborting." << endl ;
    throw "Gdb failed!" ;
  }
  if (lastReadLineSDoesContainRegexp("No such file or directory") ) {
    cerr << "In MonitorGDB::gdbSource command is " << commandString << endl ;
    cerr << "Gdb failed to find the file. Aborting." << endl ;
    throw "Gdb failed!" ;
  }

  restoreTracingMode(interceptedTracingMode) ;

} // EndMethod MonitorGDB::gdbSource

/**
 * allocate a new string (char *) and fills it with the
 * passed string arguments.
 */
char* MonitorGDB::createConcatenatedString( const char * begin  , 
					    const char * middle ,
					    const char * end ) 
{
  
  int lenBegin       = strlen( begin  ) ;
  int lenMiddle      = strlen( middle ) ;
  int completeLength = strlen( end    ) + lenBegin + lenMiddle ;

  char * completeExpression = (char*)malloc(completeLength+1) ; 
  // 1 for the terminating \0

  strcpy( completeExpression + 0                 , begin  ) ;
  strcpy( completeExpression + lenBegin          , middle ) ;
  strcpy( completeExpression + lenBegin+lenMiddle, end    ) ;

  return completeExpression ;

} // EndMethod createConcatenatedString

// -----------------------------------------------------------

char* MonitorGDB::createRegularExpForBreakpointDectection
   (const char * funcNameRegexp) 
{
//     char * beginningOfExpression = "Breakpoint [0-9]*, 0x[a-f0-9]* in " ;
//     char *       endOfExpression = " \\(\\) from /lib" ;

  char * beginningOfExpression = "Breakpoint [0-9]*, \\(0x[a-f0-9]* in \\|\\)\\(" ;
  char *       endOfExpression = "\\) ([^)]*) \\(from /lib\\|at\\)" ;
  //  char *       endOfExpression = " () from /lib" ;

  // SYMBOL_SUBSTRING_INDEX indicates the index of the submatch
  // (called substring in the regexec man pages) that corresponds to
  // the search funcNameRegexp. See the comments in
  // 'setBreakPointSOnSymbol' for more insight on "substring" indexing
  // with regexec.

#define SYMBOL_SUBSTRING_INDEX 2

   return createConcatenatedString( beginningOfExpression,
				    funcNameRegexp,
				    endOfExpression) ;

} // EndMethod MonitorGDB::createRegularExpForBreakpointDectection

// -----------------------------------------------------------

bool MonitorGDB::identifyBreakpointInLastReadLines(regmatch_t& result) {

  bool interceptedTracingMode = interceptsTracingMode() ;

  if (interceptedTracingMode) 
    cerr << "identifyBreakpointInLastReadLines" << endl ;

  char * completeExpression = 
    createRegularExpForBreakpointDectection(".*") ;

  regmatch_t matchArray[SYMBOL_SUBSTRING_INDEX+1] ;
    
  bool returnValue 
    = this->matchRegexpInStringGetSubmatch( completeExpression,
					    lastReadLineS,
					    SYMBOL_SUBSTRING_INDEX+1,
					    matchArray) ;

  result = matchArray[SYMBOL_SUBSTRING_INDEX] ;

  free(completeExpression) ;

  restoreTracingMode(interceptedTracingMode) ;

  if (interceptedTracingMode)
    cerr << "Result " << (returnValue?"YES":"NO") << endl ;

  return returnValue ;

} // EndMethod MonitorGDB::identifyBreakpointInLastReadLines


// -----------------------------------------------------------

void MonitorGDB::continueUntillBreakpointOnFunction( char * funcNameRegexp ) {

  bool interceptedTracingMode = interceptsTracingMode() ;

  if (interceptedTracingMode) 
    cerr << "continueUntillBreakpointOnFunction: " << funcNameRegexp << endl ;

  if (funcNameRegexp==NULL) {
    cerr << "MonitorGDB::continueUntillBreakpointOnFunction called with null arg." << endl ;
    throw "MonitorGDB failed!" ;
  }

  char * completeExpression = 
    createRegularExpForBreakpointDectection(funcNameRegexp) ;

  bool stopBecauseGdbSaidProgramNotBeingRun = false ;

  do    {   
    this->sendLineToChildBlocking( "continue" ) ;
    stopBecauseGdbSaidProgramNotBeingRun = gdbSaidProgramNotBeingRun() ;
  } 
  while ( ! stopBecauseGdbSaidProgramNotBeingRun && 
	  ! this->lastReadLineSDoesContainRegexp (completeExpression)) ;

  if (interceptedTracingMode)
    if (stopBecauseGdbSaidProgramNotBeingRun)
	cerr << "continueUntillBreakpointOnFunction: " 
	     << "PROGRAM HAS STOPPED !" << endl ;
      else
	cerr << "continueUntillBreakpointOnFunction: " 
	     << funcNameRegexp << " FOUND" << endl ;

  free(completeExpression) ;

  restoreTracingMode(interceptedTracingMode) ;

} // EndMethdo MonitorGDB::continueUntillBreakpointOnFunction


// -----------------------------------------------------------

void MonitorGDB::setBreakPointSOnSymbol( char * symbol ) {

  bool interceptedTracingMode = interceptsTracingMode() ;

  if (interceptedTracingMode) 
    cerr << "setBreakPointSOnSymbol: " << symbol << endl ;

  char * infoCommand 
    = createConcatenatedString( "info functions ^", symbol, "$") ;
  sendLineToChildBlocking(infoCommand) ;
  free(infoCommand) ;

  char * resultFromInfoCommand = strdup(lastReadLineS) ;

  char * completeExpression
    = createConcatenatedString( "^\\(0x[0-9a-f]*\\) *", symbol, "" ) ;

  // Quite tricky here. matchRegexpInStringFromOffsetGetSubmatch,
  // which is used in the following while loop, calls regexec of POSIX
  // interface. regexec uses an array of regmatch_t to store the
  // submatches of the regular expression. Those 'submatches' (called
  // substrings in the man pages), basically correspond to the \(...\)
  // chuncks of the regular expressions, except for currentMatch[0],
  // that corresponds to the overall matching string. Since we have
  // one \(...\) pair in the above regular expression, we need an
  // array with two positions: currentMatch[0] contains the whole
  // matching string, and currentMatch[1] contains the substring of
  // currentMatch[0] that matches the \(...\) pair.

  regmatch_t currentMatch[2] ;

  int outputIndex = 0 ;

  // WHILE LOOP ON THE LAST READ LINES

  while (outputIndex < strlen(resultFromInfoCommand)) {

    bool resultOfSearch = matchRegexpInStringGetSubmatch
      ( completeExpression, resultFromInfoCommand+outputIndex, 2, currentMatch ) ;
    if (!resultOfSearch) {
      outputIndex = strlen(resultFromInfoCommand) ;
      break ;
    } 
    // EndIf we did not find the regular expression. 
    // This means there is no address left that corresponds
    // so the searched symbol.

    char* match = createSubstringFromLastReadLineS
      ( currentMatch[1].rm_so + outputIndex ,
	currentMatch[1].rm_eo + outputIndex ) ;

    char * breakCommand 
      = createConcatenatedString( "break *", match, "" ) ;
    sendLineToChildBlocking( breakCommand ) ;
    free( breakCommand ) ;
    free( match        ) ;

    outputIndex = outputIndex + currentMatch[0].rm_eo ;

  } // EndWhile

  // END WHILE LOOP

  free(completeExpression) ;
  free(resultFromInfoCommand) ;

  restoreTracingMode(interceptedTracingMode) ;

} // EndMethod MonitorGDB::setBreakPointSOnSymbol

// -----------------------------------------------------------

bool MonitorGDB::gdbSaidProgramNotBeingRun() {
  bool interceptedTracingMode = interceptsTracingMode() ;

  bool result = 
    lastReadLineSDoesContainRegexp ("The program is not being run.") ;

  if (interceptedTracingMode) 
    cerr << "gdbSaidProgramNotBeingRun: " << (result?"YES":"NO") << endl ;

  restoreTracingMode(interceptedTracingMode) ;  

  return result ;

} // EndMethod gdbSaidProgramNotBeingRun

// -----------------------------------------------------------

void MonitorGDB::continueUntilProgramStops()
{
  bool interceptedTracingMode = interceptsTracingMode() ;
  
  while (!gdbSaidProgramNotBeingRun()) {
    //      aMonitor.sendLineToChildBlocking( "bt"); 
    sendLineToChildBlocking( "continue"); 

    if (interceptedTracingMode) { 

      // We check while the program stops. If it was because of a
      // breakpoint we print the corresponding symbol.

      regmatch_t resultOfIdentification ;
      if (identifyBreakpointInLastReadLines(resultOfIdentification)) {
	char* breakpoint 
	  = createSubstringFromLastReadLineS( resultOfIdentification.rm_so ,
					      resultOfIdentification.rm_eo ) ;
	cerr << "In continueUntilProgramStops(): Breakpoint on " << breakpoint << endl ;
	free( breakpoint ) ;
      } // EndIf a break point was found.
      else
	cerr << "In continueUntilProgramStops()" << endl ;
      
    } // EndIf if (interceptedTracingMode)

  } // EndWhile

  restoreTracingMode(interceptedTracingMode) ;  

} // EndMethod traceUntillProgramStops
